#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/base/AmpelFlexModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.03.2021
# Last Modified Date:  06.02.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.base.AmpelBaseModel import AmpelBaseModel


class AmpelFlexModel(AmpelBaseModel):
	"""
	Allows/Ignores extra kwargs parameters
	"""

	def __init__(self, **kwargs):
		super().__init__(**{k: kwargs[k] for k in kwargs if k in self.get_model_keys()})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/abstract/AbsT1RetroCombineUnit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.06.2021
# Last Modified Date:  17.06.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Union
from collections.abc import Iterable
from ampel.types import ChannelId, DataPointId
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.LogicalUnit import LogicalUnit
from ampel.content.DataPoint import DataPoint
from ampel.struct.T1CombineResult import T1CombineResult


class AbsT1RetroCombineUnit(AmpelABC, LogicalUnit, abstract=True):
	""" A unit that combines datapoints """

	debug: bool = False
	channel: ChannelId
	access: list[int | str]
	policy: list[int | str]

	@abstractmethod
	def combine(self, datapoints: Iterable[DataPoint]) -> list[list[DataPointId]] | list[T1CombineResult]:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/struct/JournalAttributes.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                15.10.2018
# Last Modified Date:  12.05.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from collections.abc import Sequence
from ampel.types import Tag
from ampel.content.JournalRecord import JournalRecord


class JournalAttributes:
	"""
	Structure potentialy used by Ampel units to customize
	the stock journal entry created after a process is run.
	"""

	__slots__ = 'tag', 'code', 'extra'

	#: code / status
	code: None | int

	#: journal entry tag(s)
	tag: None | Tag | Sequence[Tag]

	#: if provided, will be included as-is under the journal root key 'extra'
	extra: None | dict[str, Any]


	def __init__(self,
		code: None | int = None,
		tag: None | Tag | Sequence[Tag] = None,
		extra: None | dict[str, Any] = None
	) -> None:
		"""
		:param code: potential integer code of the executed process
		:param tag: journal entry tag(s)
		:param extra: optional dict, which if provided, will be included 'as is' under journal root key 'extra'
		Note regarding the type of extra: as of june 2020, mypy does not support recursive / json type
		-> https://github.com/python/mypy/issues/731
		"""
		self.tag = tag
		self.extra = extra
		self.code = code


	def dict(self) -> JournalRecord:
		d: JournalRecord = {}
		if self.tag:
			d['tag'] = self.tag
		if self.code:
			d['code'] = self.code
		if self.extra:
			d['extra'] = self.extra
		return d
			

	def into(self, prime: JournalRecord) -> JournalRecord:

		if self.code:
			prime['code'] = self.code

		if self.tag:
			prime['tag'] = self.tag

		if self.extra:
			if 'extra' in prime:
				prime['extra'] = self.extra | prime['extra']
			else:
				prime['extra'] = self.extra

		return prime



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/struct/StockAttributes.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                02.09.2021
# Last Modified Date:  02.09.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Sequence
from ampel.types import Tag
from ampel.struct.JournalAttributes import JournalAttributes


class StockAttributes:
	"""
	Structure potentialy used by Ampel units to customize
	journal entries along with stock tags or names.
	"""

	__slots__ = 'journal', 'tag', 'name'

	#: Journal entry customization (mandatory)
	journal: JournalAttributes

	#: Stock document tag(s)
	tag: None | Tag | Sequence[Tag]

	#: stock external name(s) / reference id(s)
	name: None | str | Sequence[str]


	def __init__(self,
		journal: JournalAttributes,
		tag: None | Tag | Sequence[Tag] = None,
		name: None | str | Sequence[str] = None
	) -> None:
		"""
		:param journal: Journal entry customization
		:param tag: stock document tag(s)
		:param name: stock external name(s) / reference id(s)
		"""
		self.journal = journal
		self.tag = tag
		self.name = name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/util/compression.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                29.06.2021
# Last Modified Date:  20.04.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import zipfile
from io import BytesIO
from typing import Literal

TCompression = Literal['ZIP_DEFLATED', 'ZIP_LZMA', 'ZIP_BZIP2']

def compress(
	payload: bytes,
	filename: str,
	alg: TCompression = "ZIP_DEFLATED",
	compression_level: int = 9
) -> bytes:

	outbio, zf = _new_zip_file(alg, compression_level)
	zf.writestr(filename, payload)
	zf.close()
	return outbio.getvalue()


def compress_many(
	arg: dict[str, bytes],
	alg: TCompression = "ZIP_DEFLATED",
	compression_level: int = 9
) -> bytes:

	outbio, zf = _new_zip_file(alg, compression_level)
	for k, v in arg.items():
		zf.writestr(k, v)
	zf.close()

	return outbio.getvalue()


def _new_zip_file(
	alg: TCompression = "ZIP_DEFLATED",
	compression_level: int = 9
) -> tuple[BytesIO, zipfile.ZipFile]:

	outbio = BytesIO()
	return outbio, zipfile.ZipFile(
		outbio, "w", getattr(zipfile, alg), False,
		compresslevel = compression_level
	)


def decompress(arg: bytes) -> bytes:
	bio = BytesIO()
	bio.write(arg)
	zf = zipfile.ZipFile(bio)
	file_name = zf.namelist()[0]
	return zf.read(file_name)


def decompress_str(arg: bytes) -> str:
	return str(decompress(arg), "utf8")


def decompress_many(arg: bytes) -> dict[str, bytes]:
	bio = BytesIO()
	bio.write(arg)
	zf = zipfile.ZipFile(bio)
	return {file_name: zf.read(file_name) for file_name in zf.namelist()}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/model/UnitModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                26.09.2019
# Last Modified Date:  15.09.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, Generic, Literal
from ampel.types import T
from ampel.base.AmpelGenericModel import AmpelGenericModel


class UnitModel(AmpelGenericModel, Generic[T]):
	"""
	Specification of a processing unit.
	Note: generic parametrization allows to constrain unit ids (ex: UnitModel[Literal['T2SNCosmo']])
	"""

	#: Name of ampel unit class
	unit: T

	#: - None: no config (use class defaults)
	#: - dict: config 'as is'
	#: - str: a corresponding alias key in the AmpelConfig must match the provided string
	#: - int: used internally for T2 units, a corresponding int key (AmpelConfig, base key 'confid') must match the provided integer
	config: None | int | str | dict[str, Any] = None

	secrets: None | dict[str, Any] = None

	#: Values to override in the config
	override: None | dict[str, Any] = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/abstract/AbsTiedStateT2Unit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                11.03.2020
# Last Modified Date:  28.09.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Generic
from collections.abc import Sequence
from ampel.types import UBson, T
from ampel.struct.UnitResult import UnitResult
from ampel.view.T2DocView import T2DocView
from ampel.base.decorator import abstractmethod
from ampel.content.T1Document import T1Document
from ampel.content.DataPoint import DataPoint
from ampel.abstract.AbsTiedT2Unit import AbsTiedT2Unit
from ampel.model.StateT2Dependency import StateT2Dependency


class AbsTiedStateT2Unit(Generic[T], AbsTiedT2Unit, abstract=True):
	"""
	A T2 unit bound to a :class:`~ampel.content.T1Document.T1Document` (state of a stock),
	as well as the results of other T2 units
	"""

	t2_dependency: Sequence[StateT2Dependency[T]]

	@abstractmethod
	def process(self,
		compound: T1Document,
		datapoints: Sequence[DataPoint],
		t2_views: Sequence[T2DocView]
	) -> UBson | UnitResult:
		"""
		Returned object should contain computed science results to be saved into the DB.

		.. note:: the returned dict must have only string keys and be BSON-encodable
		"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/abstract/AbsT3ReviewUnit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                23.02.2018
# Last Modified Date:  27.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import TypeVar, Generic
from collections.abc import Generator
from ampel.types import UBson, T3Send
from ampel.view.SnapView import SnapView
from ampel.view.T3Store import T3Store
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.LogicalUnit import LogicalUnit
from ampel.struct.UnitResult import UnitResult

T = TypeVar("T", bound=SnapView)


class AbsT3ReviewUnit(Generic[T], AmpelABC, LogicalUnit, abstract=True):
	""" Generic abstract class for T3 units receiving a SnapView generator """

	# avoid introspection at run-time
	_View: type[T] = SnapView # type: ignore[assignment]

	@abstractmethod
	def process(self,
		gen: Generator[T, T3Send, None],
		t3s: T3Store
	) -> UBson | UnitResult:
		"""
		T3 units receive SnapView instances (or subclasses of) via a generator.
		The method gen.send(...) applies a modification to the last view yielded by the generator
		unless a stock id is provided via a tuple.
		Use gen.send(JournalAttributes) to customize the journal of the stock document associated with the view.
		Use gen.send(StockAttributes) to customize the stock tags or name (or journal) associated with the view.
		Use gen.send((StockId, JournalAttributes)) to customize the journal for a view other than the most recent,
		e.g. when processing in batches.
		The content of the t3 store is dependent on:
		- the configuration of the 'include' option of the underlying t3 process
		- previously run t3 units if the option 'propagate' is activated
		"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/enum/EventCode.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                04.03.2021
# Last Modified Date:  25.07.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from enum import IntEnum

# flake8: noqa (E221)
class EventCode(IntEnum):
	"""
	Potential state of :class:`~ampel.content.EventDocument.EventDocument`
	"""

	OK                        = 0
	UNDEFINED                 = 1
	PRE_CHECK_EXIT            = 2
	ABORTED	                  = 3

	ERROR                     = -1
	EXCEPTION                 = -2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/base/AmpelABC.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                27.12.2017
# Last Modified Date:  17.02.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import inspect
from typing import Callable


class AmpelABC:
	"""
	This class resembles python's standart ABC module (Abstract Base Class) but can additionaly
	check method signatures. As a consequence, a subclass that inherits AmpelABC, will not be able to
	implement methods declared as abstract by the parent class using different method arguments.
	Notes:
	- Multi-level and multiple inheritance are supported.
	- Overriding abstractmethod is supported (if subclass itself is abstract)
	- This class relies on the module 'inspect'
	- Setting AmpelABC._abcheck = False deactivates all checks
	- If a sub-classes implements __init_subclass__ for some reason, \
		super().__init_subclass__(**kwargs) must be called within the (class) method.
	"""

	_abcheck = True

	@classmethod
	def __init_subclass__(cls, abstract: bool = False, **kwargs) -> None:
		"""
		Creates the corresponding class
		Note: all checks can be deactivated by setting AmpelABC._abcheck = False

		:raises NotImplementedError: if an abstract method is not implemented by the child class
		:raises TypeError: if decorator flag check_signature was set for an abstract method but \
		the corresponding method signatures differ implementation by the sub class \
		:raises ValueError: if decorator flag check_super_call was set and the corresponding method \
		omits to call to the super method.
		"""

		# https://github.com/python/mypy/issues/5887
		super().__init_subclass__(**kwargs) # type: ignore

		# Class is abstract
		if abstract:
			setattr(cls, '__new__', _raise_error)
			if cls._abcheck:
				cls._check_methods(cls, "force_check")
		else:
			setattr(cls, '__new__', __std_new__)
			for method_name, method in cls.__dict__.items():
				if hasattr(method, "abstract_method"):
					raise ValueError(
						f"Method {method_name} cannot be marked as abstract "
						f"since {cls.__name__} is not an abstract class"
					)

			if cls._abcheck:
				cls._check_methods(cls, "abstract_method")
				cls._check_methods(cls, "default_method")


	@staticmethod
	def _check_methods(Klass: type, func_attr: str) -> None:

		# Gather abstract methods (marked by the decorator @abstractmethod)
		abs_methods = {
			method_name: (base_cls, method)
			for base_cls in reversed(Klass.mro())
				for method_name, method in base_cls.__dict__.items()
					if hasattr(method, func_attr)
		}

		# Check implementation
		for method_name, value in abs_methods.items():

			if value[0] == Klass:
				continue

			# Check if method was implemented by child
			func = getattr(Klass, method_name)
			if func.__qualname__.split(".")[0] == value[0].__name__:

				# default implementation available
				if hasattr(value[1], "default_method"):
					continue
				raise NotImplementedError(
					f"Class {Klass.__name__} must implement abstract method "
					f"'{method_name}' defined in class {value[0].__name__}"
				)

			if hasattr(value[1], "check_super_call"):
				src = inspect.getsource(func)
				if not (f"super().{method_name}" in src or f".{method_name}(self" in src):
					raise ValueError(
						f"Method {method_name} from class {Klass.__name__} "
						f"must call super().{method_name}(...)"
					)

			# No signature check for methods allowing variable arguments
			if hasattr(value[1], "var_args"):
				continue

			# Get implemented method signature
			if method_name in Klass.__dict__:
				impl = Klass.__dict__[method_name]
			else:
				for K in reversed(Klass.mro()):
					if method_name in K.__dict__:
						impl = K.__dict__[method_name]
						break

			# Get abstract method signatures
			if isinstance(value[1], classmethod):
				abstract_sig = inspect.signature(value[1].__func__)
				if isinstance(impl, classmethod):
					impl = impl.__func__
				else:
					raise TypeError(f"@classmethod missing for '{Klass.__name__}.{method_name}(...)'")
			elif isinstance(value[1], staticmethod):
				abstract_sig = inspect.signature(value[1].__func__)
				if isinstance(impl, staticmethod):
					impl = impl.__func__
				else:
					raise TypeError(f"@staticmethod missing for '{Klass.__name__}.{method_name}(...)'")
			else:
				abstract_sig = inspect.signature(value[1])

			impl_sig_keys = list(inspect.signature(impl).parameters.keys())

			# Manually at cls for bound methods
			if (inspect.ismethod(impl)):
				impl_sig_keys.insert(0, 'cls')

			if (
				len(abstract_sig.parameters) != len(impl_sig_keys) or
				hasattr(value[1], "check_signature") and
				# important cast because odict_keys(['a', 'b']) == odict_keys(['b', 'a']) is True
				list(abstract_sig.parameters.keys()) != list(impl_sig_keys)
			):
				raise TypeError(
					f"Wrong method signature. Please change the arguments of method "
					f"'{method_name}' to match those defined by the corresponding "
					f"abstract method in class {value[0].__name__}\n"
					f"Required: {list(abstract_sig.parameters.keys())}\n"
					f"Implemented: {impl_sig_keys}\n"
				)


def _raise_error(cls, *args, **kwargs) -> None:
	"""
	Abstract classes cannot be instantiated
	:raises: TypeError
	"""
	raise TypeError(
		f"Class {cls.__name__} is abstract and can thus not be instantiated"
	)


def __std_new__(mcs, *arg, **kwargs) -> Callable:
	""" Standard class creation """
	cls = None
	for el in mcs.__mro__:
		# stop at first built-in type ('object' usually, but it can be 'dict'
		# for example if # the abstract subclass inherits from a python primitive type.
		# Avoids this kind of error: TypeError: object.__new__(MyDict) is not safe, use dict.__new__()
		if '__new__' in el.__dict__ and type(el.__dict__['__new__']).__name__ == "builtin_function_or_method":
			cls = el
			break
	return cls.__new__(mcs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/view/T2DocView.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                10.02.2021
# Last Modified Date:  01.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from datetime import datetime
from typing import Any, Literal, overload
from collections.abc import Sequence
from ampel.types import StockId, UBson, T2Link, Tag, T
from ampel.content.MetaRecord import MetaRecord
from ampel.content.T2Document import T2Document
from ampel.config.AmpelConfig import AmpelConfig

TYPE_POINT_T2 = 0 # linked with datapoints, that is with tier 0
TYPE_STATE_T2 = 1 # linked with compounds, that is with tier 1
TYPE_STOCK_T2 = 3 # linked with stock document


class T2DocView:
	"""
	View of a given T2Document (with unique stock id).
	A t2 view contains read-only information from a T2Document
	and provides convenience methods to access it.
	"""

	__slots__ = 'unit', 'config', 'link', 'stock', 'tag', 'code', 'meta', 't2_type', 'body'

	stock: StockId | Sequence[StockId]
	unit: int | str
	config: None | dict[str, Any]
	link: T2Link
	tag: Sequence[Tag]
	code: int
	t2_type: int
	meta: Sequence[MetaRecord]
	body: None | Sequence[UBson]


	@classmethod # Static ctor
	def of(cls, doc: T2Document, conf: None | AmpelConfig = None) -> "T2DocView":
		"""
		We might want to move this method elsewhere in the future
		"""

		if conf:
			t2_unit_info = conf.get(f'unit.{doc["unit"]}', dict)
			if not t2_unit_info:
				raise ValueError(f'Unknown T2 unit {doc["unit"]}')

			if 'AbsStockT2Unit' in t2_unit_info['base']:
				t2_type: int = TYPE_STOCK_T2
			elif 'AbsPointT2Unit' in t2_unit_info['base']:
				t2_type = TYPE_POINT_T2
			else: # quick n dirty
				t2_type = TYPE_STATE_T2
		else:
			t2_type = -1

		return cls(
			stock = doc['stock'],
			unit = doc['unit'],
			t2_type = t2_type,
			link = doc['link'],
			tag = doc.get('tag', []),
			code = doc['code'],
			meta = doc.get('meta', []),
			body = doc.get('body'),
			config = conf.get(f'confid.{doc["config"]}', dict) if (conf and doc['config']) else None
		)


	def __init__(self,
		stock: StockId | Sequence[StockId],
		unit: int | str,
		link: T2Link,
		tag: Sequence[Tag],
		code: int,
		t2_type: int,
		meta: Sequence[MetaRecord],
		config: None | dict[str, Any] = None,
		body: None | Sequence[UBson] = None
	):
		sa = object.__setattr__
		sa(self, 'stock', stock)
		sa(self, 'unit', unit)
		sa(self, 'link', link)
		sa(self, 'tag', tag)
		sa(self, 'code', code)
		sa(self, 'body', body)
		sa(self, 'meta', meta)
		sa(self, 'config', config)
		sa(self, 't2_type', t2_type)


	def __setattr__(self, k, v):
		raise ValueError("T2DocView is read only")


	def __delattr__(self, k):
		raise ValueError("T2DocView is read only")


	def __reduce__(self):
		return (
			type(self), (
				self.stock, self.unit, self.link, self.tag, self.code,
				self.body, self.meta, self.config, self.t2_type
			)
		)


	def serialize(self) -> dict[str, Any]:
		return {k: getattr(self, k) for k in self.__slots__ if k != '_frozen'}


	def has_content(self) -> bool:
		return True if self.body else False


	def get_payload(self, code: None | int = None) -> UBson:
		"""
		:returns: the content of the last array element of body associated with a meta code >= 0 or equals code arg.
		"""
		if not self.body:
			return None

		idx = len(
			[
				el for el in self.meta
				if el['tier'] == 2 and
				(el['code'] >= 0 if code is None else el['code'] == code)
			]
		) - 1

		if idx == -1:
			return None

		# A manual/admin $unset: {body: 1} was used to delete bad data
		if idx > len(self.body) - 1:
			idx = len(self.body) - 1

		return self.body[idx] if idx >= 0 else None


	def is_point_type(self) -> bool:
		return self.t2_type == TYPE_POINT_T2


	def is_stock_type(self) -> bool:
		return self.t2_type == TYPE_STOCK_T2


	def is_state_type(self) -> bool:
		return self.t2_type == TYPE_STATE_T2


	def get_value(self,
		key: str,
		rtype: type[T], *,
		code: None | int = None,
	) -> None | T:
		"""
		:returns: the value of a given key from the content of the last array element of body
		associated with a meta code >= 0 or equals code arg

		Examples:
		get_value("fit_result", dict)
		"""
		r = self.get_payload(code)
		if isinstance(r, dict) and key in r:
			return r[key]
		return None


	@overload
	def get_ntuple(self,
		key: tuple[str, ...], rtype: type[T], *,
		no_none: Literal[True], require_all_keys: bool, code: None | int
	) -> None | tuple[T, ...]:
		...

	@overload
	def get_ntuple(self,
		key: tuple[str, ...], rtype: type[T], *,
		no_none: Literal[False], require_all_keys: bool, code: None | int
	) -> None | tuple[None | T, ...]:
		...

	def get_ntuple(self,
		key: tuple[str, ...],
		rtype: type[T], *,
		no_none: bool = False,
		require_all_keys: bool = True,
		code: None | int = None,
	) -> None | tuple[T, ...] | tuple[None | T, ...]:
		"""
		Returns a tuple of n values from the content of the last array element of body
		associated with a meta code >= 0 or equals code arg

		Examples:

		In []: tv = T2DocView(
			body=[{'a': 1, 'b': None, 'c': 3}], unit=None, link=None, tag=None,
			code=0, t2_type=0, meta=[{'tier': 2, 'code':0}], stock=0
		)

		In []: tv.get_ntuple(("a", "b"), int, no_none = False)
		Out[]: (1, None)

		In []: tv.get_ntuple(("a", "b"), int, no_none = True)
		Out[]: None

		In []: tv = T2DocView(body=[{'a': 1, 'c': 3}], ...)

		In []: tv.get_ntuple(("a", "b"), int)
		Out[]: None

		In []: tv.get_ntuple(("a", "b"), int, require_all_keys = False)
		Out[]: (1, None)

		In []: tv.get_ntuple(("a", "b"), int, require_all_keys = False, no_none = True)
		Out[]: None
		"""

		r = self.get_payload(code)

		if isinstance(r, dict):
			if r and (sks := r.keys() & key):

				if require_all_keys and len(sks) != len(key):
					return None

				t = tuple(
					r[k] if (k in r and isinstance(r[k], rtype)) else None
					for k in key
				)

				return None if (no_none and None in t) else t # type: ignore[return-value]

		return None


	@overload
	def get_time_created(self, to_string: Literal[False]) -> None | float:
		...
	@overload
	def get_time_created(self, to_string: Literal[True]) -> None | str:
		...
	def get_time_created(self, to_string: bool = False) -> None | float | str:

		ts = self.meta[0]['ts']

		if to_string:
			return datetime.fromtimestamp(ts).strftime('%d/%m/%Y %H:%M:%S')

		return ts


	@overload
	def get_time_updated(self, to_string: Literal[False]) -> None | float:
		...
	@overload
	def get_time_updated(self, to_string: Literal[True]) -> None | str:
		...
	def get_time_updated(self, to_string: bool = False) -> None | float | str:

		if not self.has_content():
			return None

		ts = self.meta[-1]['ts']
		if to_string:
			return datetime.fromtimestamp(ts).strftime('%d/%m/%Y %H:%M:%S')

		return ts



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/enum/DocumentCode.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.12.2017
# Last Modified Date:  17.05.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from enum import IntEnum

# flake8: noqa (E221)
class DocumentCode(IntEnum):
	"""
	Potential code of:
		- :class:`~ampel.content.T1Document.T1Document`
		- :class:`~ampel.content.T2Document.T2Document`
		- :class:`~ampel.content.T3Document.T3Document`

	Negative code values are reserved for the ampel system
	(typically set by ingesters or processors).
	They hint that the associated ampel document is either in an
	erroneous state or not yet processed (pending state).

	Database queries can thus use {'code': {'$gte': 0}}
	to ensure the retrieval of valid processed data only.

	Codes are not combinable (IntEnum != IntFlag) meaning
	using DocumentCode.T2_UNKNOWN_CONFIG|DocumentCode.ERROR is not allowed.
	"""

	# General
	OK                        = 0
	NEW                       = -1
	ERROR                     = -2
	INTERNAL_ERROR            = -3
	EXCEPTION                 = -4
	RUNNING                   = -5
	RERUN_REQUESTED           = -6
	TOO_MANY_TRIALS           = -7

	# T1
	T1_NEW_PRIO               = -1000  # For now, std ingesters do not support this
	T1_UNKNOWN_CONFIG         = -1001

	# T2
	T2_NEW_PRIO               = -2000  # For now, std ingesters do not support this
	T2_PENDING_DEPENDENCY     = -2001
	T2_QUEUED                 = -2002
	T2_EXPORTED               = -2003
	T2_UNKNOWN_LINK           = -2004 # might be an ingester bugs
	T2_UNKNOWN_CONFIG         = -2005
	T2_MISSING_DEPENDENCY     = -2006
	T2_UNEXPECTED_DEPENDENCY  = -2007
	T2_MISSING_INFO           = -2008
	T2_OUTDATED_CODE          = -2009

	# T3
	T3_CONTEXT_ERROR          = -3000 # error occured in context stage
	T3_SELECT_ERROR           = -3001 # error occured in select stage
	T3_LOAD_ERROR             = -3002 # error occured in load stage
	T3_COMPLEMENT_ERROR       = -3003 # error occured in complement stage
	T3_RUN_ERROR              = -3004 # error occured in run stage



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/model/StateT2Dependency.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                06.02.2021
# Last Modified Date:  28.09.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Generic
from ampel.types import T
from ampel.model.UnitModel import UnitModel
from ampel.model.DPSelection import DPSelection


class StateT2Dependency(UnitModel[T], Generic[T]):
	"""
	Used to specify how "tied state t2" units should select the associated required t2 document.

	T2Processor needs to retrieve the T2Records of units tied with this unit.
	If link_override is unspecified, t2 dependencies are resolved - for each StateT2Dependency -
	using the db match query: {unit: <unit_name>, link: <same link as root doc>}.

	This behavior is overridable/customizable via 't2_dependency->link_override' (t2 config dict).
	'link_override' allows to link a state T2 with a different value than the one registered as 'link' in the T2Record.

	For example, link_override enables to tie a state T2 with the result of a point t2
	(the returned link could be the id of the first datapoint contained in the compound)
	"""

	#: - None: the state associated with the root tied state T2 will be used (value of 'link' in t2 doc)
	#: - DPSelection: allows tied state T2 units to be bound with point t2 units.
	link_override: None | DPSelection



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/types.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                02.12.2019
# Last Modified Date:  31.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import ValuesView, KeysView, Sequence
from typing import Any, Union, TypeVar, get_origin, TYPE_CHECKING, Annotated, _GenericAlias # type: ignore[attr-defined]

if TYPE_CHECKING:
	from ampel.base.AmpelBaseModel import AmpelBaseModel
	from ampel.struct.JournalAttributes import JournalAttributes
	from ampel.struct.StockAttributes import StockAttributes

do_type_check = True

TRACELESS = -1
T = TypeVar('T')
Traceless = Annotated[T, TRACELESS]

StockId = Union[int, bytes, str]
ChannelId = Union[int, str]
DataPointId = int
Tag = Union[int, str]
UnitId = str
JDict = dict[str, Any] # JSON dict

T2Link = Union[StockId, DataPointId, int]
T3Send = Union['JournalAttributes', 'StockAttributes', tuple[StockId, 'StockAttributes']]

UBson = Union[None, str, int, float, bool, bytes, list[Any], dict[str, Any]]
TBson = TypeVar("TBson", str, int, float, bool, bytes, list, dict)
ubson = (str, int, float, bool, bytes, list, dict)
StrictIterable = Union[list[T], set[T], tuple[T], frozenset[T], ValuesView[T], KeysView[T]]
strict_iterable = (list, tuple, set, frozenset, ValuesView, KeysView)

OneOrMany = Union[Sequence[T], T] # please keep this order


def check_class(Klass: type, class_type: Union[type["AmpelBaseModel"], tuple[type["AmpelBaseModel"], ...], _GenericAlias]) -> None:
	""" :raises: ValueError """
	if isinstance(class_type, _GenericAlias):
		class_type = get_origin(class_type)
	if not issubclass(Klass, class_type):
		raise ValueError(f"{Klass} is not a subclass of {class_type}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/content/JournalRecord.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                12.12.2019
# Last Modified Date:  05.09.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Literal, Any, TypedDict
from collections.abc import Sequence
from ampel.types import ChannelId, Tag


class JournalRecord(TypedDict, total=False):
	"""
	A record of activity on a stock document.

	.. seealso:: :class:`~ampel.struct.JournalAttributes.JournalAttributes`
	"""

	#: Tier of the associated process
	tier: Literal[-1, 0, 1, 2, 3]

	#: UNIX epoch of the activity
	ts: int | float

	#: Channels associated with the activity
	channel: ChannelId | Sequence[ChannelId]

	#: Name of the associated process
	process: int | str

	#: Free-form labels
	tag: Tag | Sequence[Tag]

	#: Run(s) associated with this record
	run: int

	#: Status code of the associated process
	code: int

	#: Action code(s) built from :class:`~ampel.enum.JournalActionCode.JournalActionCode`
	action: int

	#: Duration of the process
	duration: int | float

	#: Trace ids
	traceid: dict[str, int]

	#: id of the unit associated with this record
	unit: int | str

	#: id of the document associated with the invocation
	doc: int | bytes

	#: Free-form information
	extra: dict[str, Any]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/protocol/LoggerProtocol.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                27.09.2018
# Last Modified Date:  18.12.2020
# Last Modified By:    Jakob van Santen <jakob.van.santen@desy.de>

from typing import Any, Protocol, runtime_checkable, TYPE_CHECKING

if TYPE_CHECKING:
	from ampel.protocol.LoggingHandlerProtocol import LoggingHandlerProtocol

@runtime_checkable
class LoggerProtocol(Protocol):

	def error(self, msg: str | dict[str, Any], *args,
		exc_info: None | Exception = None,
		extra: None | dict[str, Any] = None,
	) -> None:
		...

	def warn(self, msg: str | dict[str, Any], *args,
		extra: None | dict[str, Any] = None,
	) -> None:
		...

	def info(self, msg: None | str | dict[str, Any], *args,
		extra: None | dict[str, Any] = None,
	) -> None:
		...

	def debug(self, msg: None | str | dict[str, Any], *args,
		extra: None | dict[str, Any] = None,
	):
		...

	def log(self,
		lvl: int, msg: None | str | dict[str, Any], *args,
		exc_info: None | bool | Exception = None,
		extra: None | dict[str, Any] = None,
	):
		...

	def addHandler(self, handler: "LoggingHandlerProtocol"):
		...

	def removeHandler(self, handler: "LoggingHandlerProtocol"):
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/cli/ArgParserBuilder.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.03.2021
# Last Modified Date:  23.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser


class ArgParserBuilder:
	"""
	Creates one or many instances of AmpelArgumentParser
	"""

	def __init__(self, op: None | str = None) -> None:
		"""
		"""
		self.parsers: dict[str, AmpelArgumentParser] = {}
		self._op = op


	def add_parser(self, name: str, hlp: dict[str, str]):
		self.parsers[name] = AmpelArgumentParser(ampel_op=self._op)
		self.parsers[name].set_help_descr(hlp)
		setattr(self.parsers[name], '_sub_op', name)


	def add_parsers(self, names: list[str], hlp: dict[str, str]):
		for el in names:
			self.add_parser(el, hlp)


	def get(self) -> dict[str, AmpelArgumentParser]:
		""" Returns parsers created during the build procedure """
		return self.parsers


	def add_group(self, target: str, title: str, **kwargs):
		"""
		:param target: see add_arg(...) docstring
		"""
		parsers, group = self.get_targets(target, False)
		for p in parsers:
			p.groups[group] = p.add_argument_group(group, **kwargs)
			p.groups[group].title = title


	def set_group_defaults(self, target: str, **kwargs):
		"""
		:param target: see add_arg(...) docstring
		"""
		parsers, group = self.get_targets(target, True)
		for p in parsers:
			p.groups[group].set_defaults(**kwargs)


	def add_description(self, target: str, descr: None | str | list[str]) -> None:
		"""
		Calls method 'add_description' (see AmpelArgumentParser docstring) for the selected parser(s).
		:param target: see add_arg(...) docstring
		"""

		parsers, _ = self.get_targets(target, False)
		for p in parsers:
			p.add_description(descr)


	def add_arg(self, target: str, name: str, **kwargs):
		"""
		:param target:
		- str:
			ex: add_arg('optional', 'verbose')
			Adds the argument to the provided group ('target' value) of all underlying parsers.
			For example, if two sub-parsers exist (say one for operation 'show' and one for 'write'),
			both subparsers will feature the argument 'verbose'
		- str.str:
			ex: add_arg('write.required', 'out-file')
			The first part of the string, delimited by '.', targets a sub-parser by name.
			The second part targets an argument group (by name) of the specified subparser(s).
			In the example above, the argument 'out-file' will be added to the group
			with name 'required' of sub-parser with name 'write'.
		- all.str:
			ex: add_arg('all.required', 'out-file'). Same as the first case (str), only explicit
		- str|str.str:
			ex: add_arg('write|show|tail.optional', 'to-json')
			Multiple sub-parsers can be targeted using '|'.
			In the example above, the argument 'to-json' will be added to the
			argument group 'optional' of the sub-parser of the 'write', 'show' and 'tail' sub-operations
		"""

		parsers, group = self.get_targets(target, True)
		for p in parsers:
			p.add_arg(name, group, **kwargs)


	def add_x_args(self, target: str, *args: dict[str, Any]) -> None:
		"""
		Calls method 'add_x_args' (see AmpelArgumentParser docstring) for the selected parser(s).
		:param target: see add_arg(...) docstring
		"""
		parsers, group = self.get_targets(target, True)
		for p in parsers:
			p.add_x_args(group, *args)


	def get_targets(self, target: str, check_group: bool = True) -> tuple[list[AmpelArgumentParser], str]:
		"""
		:param target: see add_arg(...) docstring
		"""
		if "." in target:
			sp = target.split(".")
			parsers = self._get_parsers(sp[0])
			group = sp[1]
		else:
			parsers = self._get_parsers("all")
			group = target

		if check_group:
			for p in parsers:
				if group not in p.groups:
					raise ValueError(
						f"Group '{group}' unknown in parser with name "
						f"\'{getattr(p, '_sub_op')}\', please create it first\n"
					)

		return parsers, group


	def _get_parsers(self, arg: str) -> list[AmpelArgumentParser]:
		"""
		:param arg: string containing one or more parser names.
		Multiple parser names mst be separated by the character "|".
		Unlike parameter 'target' of add_arg(...), not dot ('.') can be used with 'arg'.
		"""

		if "|" in arg:
			pp = arg.split("|")
			for el in pp:
				if el not in self.parsers:
					raise ValueError(f"Unknown parser: '{el}', please create it using add_parser")

			ret1: list[AmpelArgumentParser] = [
				v for k, v in self.parsers.items()
				if k in pp
			]

		elif arg == "all":
			ret1 = list(self.parsers.values())

		else:
			if arg not in self.parsers:
				raise ValueError(f"Unknown parser: '{arg}', please create it using add_parser")
			ret1 = [self.parsers[arg]]

		return ret1


	def create_logic_args(self,
		target: str, name: str, descr: str, metavar: str = "#",
		required: bool = False, pos: None | int = None,
		ref: None | str = None, excl: bool = False,
		json: bool = True, **kwargs
	) -> None:
		"""
		Calls method 'create_logic_args' (see AmpelArgumentParser docstring) for the selected parser(s).
		:param target: see add_arg() docstring
		"""
		parsers, group = self.get_targets(target, True)
		for p in parsers:
			p.create_logic_args(name, group, descr, metavar, required, pos, ref, excl, json)


	def add_note(self, parsers: str, note: str, pos: None | int = None, ref: None | str = None) -> None:
		"""
		Calls method 'add_note' (see AmpelArgumentParser docstring) for the selected parser(s).
		:param parsers: see _get_parsers() docstring
		"""
		for p in self._get_parsers(parsers):
			p.add_note(note, pos, ref)


	def add_all_note(self, note: str, pos: None | int = None, ref: None | str = None) -> None:
		"""
		Calls method 'add_note' (see AmpelArgumentParser docstring) of all the underlying parsers.
		"""
		for p in self.parsers.values():
			p.add_note(note, pos, ref)


	def add_example(self,
		parsers: str, ex: str, prepend=None, append="",
		auto_strip: bool = True, ref: None | str = None
	) -> None:
		"""
		Calls method 'add_example' (see AmpelArgumentParser docstring) for the selected parser(s).
		:param parsers: see _get_parsers() docstring
		"""
		for p in self._get_parsers(parsers):
			p.add_example(
				ex,
				prepend or f"ampel {self._op or ''} {getattr(p, '_sub_op', '')} ",
				append, auto_strip, ref
			)


	def add_all_example(self,
		ex: str, prepend="ampel ", append="", auto_strip: bool = True, ref: None | str = None
	) -> None:
		"""
		Calls method 'add_example' (see AmpelArgumentParser docstring) of all the underlying parsers.
		"""
		for p in self.parsers.values():
			p.add_example(ex, prepend, append, auto_strip, ref)


	def hint_query_logic(self, parsers: str, pos: None | int = None, ref: None | str = None):
		for p in self._get_parsers(parsers):
			p.hint_query_logic(pos, ref)


	def hint_all_query_logic(self, pos: None | int = None, ref: None | str = None):
		for p in self.parsers.values():
			p.hint_query_logic(pos, ref)


	def hint_time_format(self, parsers: str, pos: None | int = None, ref: None | str = None):
		for p in self._get_parsers(parsers):
			p.hint_time_format(pos, ref)


	def hint_all_time_format(self, pos: None | int = None, ref: None | str = None):
		for p in self.parsers.values():
			p.hint_time_format(pos, ref)


	def hint_all_config_override(self, pos: None | int = None, ref: None | str = None):
		for p in self.parsers.values():
			p.hint_config_override(pos, ref)


	def notation_add_note_references(self):
		for p in self.parsers.values():
			p.notation_add_note_references()


	def notation_add_example_references(self):
		for p in self.parsers.values():
			p.notation_add_example_references()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/util/tag.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                25.11.2021
# Last Modified Date:  25.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.types import Tag
from typing import Literal, overload
from collections.abc import Sequence


@overload
def merge_tags(
	arg1: Tag | Sequence[Tag] | set[Tag],
	arg2: None | Tag | Sequence[Tag] | set[Tag]
) -> list[Tag]:
	...

@overload
def merge_tags(
	arg1: None | Tag | Sequence[Tag] | set[Tag],
	arg2: Tag | Sequence[Tag] | set[Tag]
) -> list[Tag]:
	...

@overload
def merge_tags(
	arg1: None | Tag | Sequence[Tag] | set[Tag],
	arg2: None | Tag | Sequence[Tag] | set[Tag]
) -> None | list[Tag]:
	...

@overload
def merge_tags(
	arg1: Tag | Sequence[Tag] | set[Tag],
	arg2: None | Tag | Sequence[Tag] | set[Tag], *,
	reduce: Literal[False]
) -> list[Tag]:
	...

@overload
def merge_tags(
	arg1: None | Tag | Sequence[Tag] | set[Tag],
	arg2: Tag | Sequence[Tag] | set[Tag], *,
	reduce: Literal[False]
) -> list[Tag]:
	...

@overload
def merge_tags(
	arg1: None | Tag | Sequence[Tag] | set[Tag],
	arg2: None | Tag | Sequence[Tag] | set[Tag], *,
	reduce: Literal[False]
) -> None | list[Tag]:
	...

@overload
def merge_tags(
	arg1: Tag | Sequence[Tag] | set[Tag],
	arg2: None | Tag | Sequence[Tag] | set[Tag], *,
	reduce: Literal[True]
) -> Tag | list[Tag]:
	...

@overload
def merge_tags(
	arg1: None | Tag | Sequence[Tag] | set[Tag],
	arg2: Tag | Sequence[Tag] | set[Tag], *,
	reduce: Literal[True]
) -> Tag | list[Tag]:
	...

@overload
def merge_tags(
	arg1: None | Tag | Sequence[Tag] | set[Tag],
	arg2: None | Tag | Sequence[Tag] | set[Tag], *,
	reduce: Literal[True]
) -> None | Tag | list[Tag]:
	...

def merge_tags(
	arg1: None | Tag | Sequence[Tag] | set[Tag],
	arg2: None | Tag | Sequence[Tag] | set[Tag], *,
	reduce: bool = True
) -> None | Tag | list[Tag]:
	""" A mypy friendly method that merges tags together """

	if arg2 is None:
		if arg1 is None:
			return None
		if isinstance(arg1, (int, str)):
			return arg1
		if reduce and len(arg1) == 1:
			if isinstance(arg1, set):
				return next(iter(arg1))
			return arg1[0]
		return arg1 if isinstance(arg1, list) else list(arg1)

	if arg1 is None:
		if arg2 is None:
			return None
		return merge_tags(arg2, arg1, reduce) # type: ignore # no idea, no time

	if isinstance(arg1, (str, int)):
		if isinstance(arg2, (str, int)):
			if arg2 != arg1:
				return [arg1, arg2]
			return arg1 if reduce else [arg1]
		else:
			if arg1 in arg2:
				return arg2 if isinstance(arg2, list) else list(arg2)
			l = list(arg2)
			l.append(arg1)
			return l

	# arg1 is a sequence (str excluded)
	else:
		if isinstance(arg2, (int, str)):
			if arg2 in arg1:
				return arg1 if isinstance(arg1, list) else list(arg1)
			l = list(arg1)
			l.append(arg2)
	
		# arg1 and arg2 are sequences (str excluded)
		else:
			l = list(
				(arg1 if isinstance(arg1, set) else set(arg1)) |
				(arg2 if isinstance(arg2, set) else set(arg2))
			)

		return l[0] if reduce and len(l) == 1 else l



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/view/SnapView.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                13.01.2018
# Last Modified Date:  09.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from datetime import datetime
from typing import Any, Literal, overload
from collections.abc import Container, Callable, Iterator, Sequence

from ampel.types import OneOrMany, StockId, T2Link, UBson, T
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.config.AmpelConfig import AmpelConfig
from ampel.content.DataPoint import DataPoint
from ampel.content.T1Document import T1Document
from ampel.view.T2DocView import T2DocView
from ampel.content.StockDocument import StockDocument
from ampel.content.LogDocument import LogDocument
from ampel.content.JournalRecord import JournalRecord
from ampel.util.freeze import recursive_freeze as rf


class SnapView:
	"""
	View of a given ampel object (with unique stock id).

	This class references various instances of objects from
	package ampel.content, originating from different ampel tiers.
	It can also contain external/composite objects embedded in the dict called 'extra',
	for example spectra or cutout images.
	The config parameter of a T3 process determines which information are included.
	Instances of this class (or of subclass such as
	:class:`~ampel.view.TransientView.TransientView`) are provided to
	:meth:`AbsT3ReviewUnit.process() <ampel.abstract.AbsT3ReviewUnit.AbsT3ReviewUnit.process>`.

	"""

	__slots__ = 'id', 'stock', 'origin', 't0', 't1', 't2', 'logs', 'extra'

	id: StockId
	stock: None | StockDocument
	origin: None | OneOrMany[int]
	t0: None | Sequence[DataPoint]
	t1: None | Sequence[T1Document]
	t2: None | Sequence[T2DocView]
	logs: None | Sequence[LogDocument]
	extra: None | dict[str, Any]


	@classmethod
	def of(cls, ab: AmpelBuffer, conf: None | AmpelConfig = None, freeze: bool = True) -> 'SnapView':

		if freeze:
			return cls(
				id = ab['id'],
				stock = rf(ab['stock']) if ab.get('stock') else None,
				origin = ab.get('origin'),
				t0 = tuple(rf(el) for el in ab['t0']) if ab.get('t0') else None, # type: ignore[union-attr]
				t1 = tuple(rf(el) for el in ab['t1']) if ab.get('t1') else None, # type: ignore[union-attr]
				t2 = tuple(T2DocView.of(rf(el), conf) for el in ab['t2']) if ab.get('t2') else None, # type: ignore[union-attr]
				logs = tuple(rf(el) for el in ab['logs']) if ab.get('logs') else None, # type: ignore[union-attr]
				extra = rf(ab['extra']) if ab.get('extra') else None
			)

		return cls(
			id = ab['id'],
			stock = ab.get('stock'),
			origin = ab.get('origin'),
			t0 = ab.get('t0'),
			t1 = ab.get('t1'),
			t2 = [T2DocView.of(el, conf) for el in ab['t2']] if ab.get('t2') else None, # type: ignore[union-attr]
			logs = ab.get('logs'),
			extra = ab.get('extra')
		)


	def __init__(self,
		id: StockId,
		stock: None | StockDocument = None,
		origin: None | OneOrMany[int] = None,
		t0: None | Sequence[DataPoint] = None,
		t1: None | Sequence[T1Document] = None,
		t2: None | Sequence[T2DocView] = None,
		logs: None | Sequence[LogDocument] = None, # Logs, if added by T3 complement stage
		extra: None | dict[str, Any] = None # Free-form information addable via instances of AbsBufferComplement
	):
		sa = object.__setattr__
		sa(self, 'id', id)
		sa(self, 'stock', stock)
		sa(self, 'origin', origin)
		sa(self, 't0', t0)
		sa(self, 't1', t1)
		sa(self, 't2', t2)
		sa(self, 'extra', extra)
		sa(self, 'logs', logs)


	def __setattr__(self, k, v):
		raise ValueError('SnapView is read only')


	def __delattr__(self, k):
		raise ValueError("SnapView is read only")


	def __reduce__(self):
		return (
			type(self),
			(
				self.id, self.stock, self.origin, self.t0,
				self.t1, self.t2, self.extra, self.logs
			)
		)


	def serialize(self) -> dict:
		return {k: getattr(self, k) for k in self.__slots__}


	# TODO: add config filter
	def get_t2_views(self,
		unit: None | str | list[str] | tuple[str, ...] = None,
		link: None | T2Link = None,
		code: None | int = None,
	) -> Iterator[T2DocView]:
		"""
		Get a subset of T2 documents.

		:param unit: limits the returned science record(s) to the one with the provided t2 unit id
		:param link: whether to return the latest science record(s) or not (default: False)

		"""

		if not self.t2:
			return None

		units: None | Container[str] = [unit] if isinstance(unit, str) else unit

		for t2v in self.t2:
			if link and t2v.link != link:
				continue
			if units and t2v.unit not in units:
				continue
			if code is not None and t2v.code != code:
				continue
			yield t2v


	def get_raw_t2_body(self,
		unit: str | list[str] | tuple[str, ...],
		link: None | T2Link = None,
		code: None | int = None,
	) -> None | Sequence[UBson]:
		"""
		:param link: restrict to a specific link
		:param code: restrict to a specific code
		"""

		# from the records that match the unit and compound selection,
		# return the last record that has a result
		for t2v in self.get_t2_views(unit, link=link, code=code):
			return t2v.body
		return None


	def get_latest_t2_body(self,
		unit: str | list[str] | tuple[str, ...],
		link: None | T2Link = None,
		code: None | int = None,
	) -> UBson:
		"""
		Get latest t2 body element from a given unit.

		:param unit_id: target unit id
		:param link_id: restrict to a specific link
		"""

		# from the records that match the unit and compound selection,
		# return the last record that has a result
		if (t2v := next(self.get_t2_views(unit, link=link, code=code), None)):
			for subrecord in reversed(t2v.body or []):
				if subrecord:
					return subrecord
		return None


	@overload
	def get_t2_body(self, unit: str | list[str] | tuple[str, ...]) -> None | dict[str, Any]:
		...
	@overload
	def get_t2_body(self, unit: str | list[str] | tuple[str, ...], ret_type: type[T]) -> None | T:
		...
	@overload
	def get_t2_body(self, unit: str | list[str] | tuple[str, ...], *, raise_exc: Literal[True]) -> T:
		...
	def get_t2_body(self,
		unit: str | list[str] | tuple[str, ...],
		ret_type: type[T] = dict, # type: ignore[assignment]
		*,
		data_slice: int = -1, # latest
		link: None | T2Link = None,
		code: None | int = None,
		raise_exc: bool = False
	) -> None | T:
		"""
		Get latest t2 body element from a given unit.
		:param ret_type: expected body element type. If isinstance check is not fullfied
		None will be returned (unless multiple unit are to be matched and another
		unit fullfill the criteria) or an exception will be raised if raise_exc is True
		:param link: restrict to a specific link
		"""
		for t2v in self.get_t2_views(unit, link=link, code=code):
			if t2v.body:
				ret = t2v.body[data_slice]
				if isinstance(ret, ret_type):
					return ret
		return None


	def get_t2_value(self,
		unit: str | tuple[str, ...],
		key: str,
		rtype: type[T], *,
		code: None | int = None
	) -> None | T:
		"""
		Examples:
		get_t2_value(("T2NedSNCosmo", "T2SNCosmo"), "fit_result", dict)

		see T2DocView.get_value(...) for more info
		"""
		for t2v in self.get_t2_views(unit):
			if (x := t2v.get_value(key, rtype, code=code)):
				return x
		return None


	@overload
	def get_t2_ntuple(self,
		unit: str | tuple[str, ...], key: tuple[str, ...], rtype: type[T], *,
		no_none: Literal[False], require_all_keys: bool = ..., code: None | int = ...
	) -> None | tuple[None | T, ...]:
		...

	@overload
	def get_t2_ntuple(self,
		unit: str | tuple[str, ...], key: tuple[str, ...], rtype: type[T], *,
		no_none: Literal[True], require_all_keys: bool = ..., code: None | int = ...
	) -> None | tuple[T, ...]:
		...

	def get_t2_ntuple(self,
		unit: str | tuple[str, ...],
		key: tuple[str, ...],
		rtype: type[T], *,
		no_none: bool = False,
		require_all_keys: bool = True,
		code: None | int = None
	) -> None | tuple[T, ...] | tuple[None | T, ...]:
		"""
		Examples:
		get_t2_ntuple("T2NedTap", ("ra", "dec", "z", "zunc"), float)
		get_t2_ntuple(("T2NedSNCosmo", "T2SNCosmo"), ("fit_result", "covariance"), dict)

		see T2DocView.get_ntuple(...) for more info
		"""

		for t2v in self.get_t2_views(unit):
			if (x := t2v.get_ntuple( # type: ignore
				key, rtype, no_none = no_none, require_all_keys = require_all_keys, code = code
			)):
				return x
		return None


	def get_journal_entries(self,
		tier: None | Literal[0, 1, 2, 3] = None,
		process_name: None | str = None,
		filter_func: None | Callable[[JournalRecord], bool] = None,
	) -> Iterator[JournalRecord]:
		"""
		Get a subset of journal entries.

		:param tier: return only journal entries associated with the given tier
		:param process_name: return only journal entries associated with a given process name
		:param latest: return only the latest entry in the journal (the latest in time)

		:returns:
			journal entries corresponding to a given tier and/or job,
			sorted by timestamp.
		"""

		if not self.stock:
			return None

		# Journal entries are sorted chronologically
		for je in self.stock['journal']:

			if tier is not None and je['tier'] != tier:
				continue

			if process_name and je['process'] != process_name:
				continue

			if filter_func and not filter_func(je):
				continue

			yield je


	def get_time_created(self,
		output: Literal['raw', 'datetime', 'str'] = 'raw'
	) -> None | float | datetime | str:

		if not self.stock:
			return None
		# Journal cannot be empty
		return self._get_time(self.stock['journal'][0], output)


	def get_time_updated(self,
		output: Literal['raw', 'datetime', 'str'] = 'raw'
	) -> None | float | datetime | str:

		if not self.stock:
			return None
		# Journal cannot be empty
		return self._get_time(self.stock['journal'][-1], output)


	@classmethod
	def _get_time(cls,
		entry: JournalRecord,
		output: None | bool | str = None
	) -> float | str | datetime:

		if output == 'raw':
			return entry['ts']

		dt = datetime.fromtimestamp(entry['ts'])

		if output == 'datetime':
			return dt

		return dt.strftime('%d/%m/%Y %H:%M:%S')


	def content_summary(self) -> str:
		return 'DP: %i, CP: %i, T2: %i' % (
			len(self.t0) if self.t0 else 0,
			len(self.t1) if self.t1 else 0,
			len(self.t2) if self.t2 else 0
		)




import inspect
import json
import bson.json_util
from importlib import import_module
from enum import IntFlag
from types import MappingProxyType
from functools import partial

def load(fileobj, ignore_missing_modules=True):
	for line in fileobj:
		yield json.loads(
			line,
			object_hook=partial(
				object_hook,
				ignore_missing_modules=ignore_missing_modules
			)
		)

class AmpelEncoder(json.JSONEncoder):
	"""
	Serialize objects in a mix of:
	a) JSONRPC 1.0-like class hinting for custom Ampel types
	b) PyMongo-provided JSON representation for BSON types
	"""
	def __init__(self, *args, lossy: bool=False, **kwargs):
		"""
		if lossy, cast mappingproxy to dict, set and tuple to list, etc
		"""
		self.lossy = lossy
		super().__init__(*args, **kwargs)
		self.bson_options = bson.json_util.CANONICAL_JSON_OPTIONS

	def default(self, obj, fallthrough=False):
		"""
		Encode selected types in jsonrpc class-hint notation
		"""
		serializer = self.get_serializer(obj)
		if serializer is not None:
			if obj.__class__.__module__ == "builtins":
				module_name = "builtins"
			else:
				module_name = inspect.getmodule(obj).__name__
			json_class = obj.__class__.__name__
			if module_name not in ['', '__main__']:
				json_class = '%s.%s' % (module_name, json_class)
			# mappingproxy lies about its type
			if type(obj) == MappingProxyType:
				json_class = 'types.MappingProxyType'
			rep = serializer(obj)
			if isinstance(rep, tuple):
				assert len(rep) == 2
				params, attrs = rep
			elif isinstance(rep, dict):
				params, attrs = [], rep
			else:
				params, attrs = rep, {}
			return {"__jsonclass__": [json_class] + self.default(params, True), **self.default(attrs, True)}
		elif hasattr(obj, 'items'):
			return {self.default(k, True): self.default(v, True) for k,v in obj.items()}
		elif isinstance(obj, list) or isinstance(obj, set) or isinstance(obj, tuple):
			return [self.default(k, True) for k in obj]
		else:
			# convert remaining Mongo types
			try:
				return bson.json_util.default(obj, self.bson_options)
			except TypeError:
				if fallthrough:
					return obj
				else:
					raise

	def get_serializer(self, obj):
		"""
		Serializers for types we want to preserve
		"""
		if hasattr(obj, 'serialize'):
			return type(obj).serialize
		elif hasattr(obj, '__dataclass_fields__'):
			return lambda x: x.__dict__
		elif isinstance(obj, IntFlag):
			return lambda x: [int(x)]
		elif self.lossy:
			return None

		# try to preserve types
		if isinstance(obj, set) or isinstance(obj, tuple):
			return lambda x: [list(x)]
		elif isinstance(obj, MappingProxyType):
			return lambda x: [dict(obj)]
		else:
			return None

def object_hook(
    dct,
    options=bson.json_util.CANONICAL_JSON_OPTIONS,
    ignore_missing_modules=True
):
	"""
	Deserialize an object serialized by AmpelEncoder
	"""
	obj = bson.json_util.object_hook(dct, options)
	if type(obj) != type(dct):
		return obj
	elif "__jsonclass__" in dct:
		ctor = dct.pop("__jsonclass__")
		parts = ctor[0].split('.')
		try:
			mod = import_module('.'.join(parts[:-1]))
		except ModuleNotFoundError:
			if ignore_missing_modules:
				return None
			else:
				raise
		klass = getattr(mod, parts[-1])
		# Here we treat the jsonrpc-style attrs as keyword args. This does not
		# necessarily conform to the 1.0 spec, but it's deprecated anyhow.
		obj = klass(*ctor[1:], **dct)
		return obj
	else:
		return dct



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/util/freeze.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                10.12.2019
# Last Modified Date:  09.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.view.ReadOnlyDict import ReadOnlyDict


def recursive_freeze(arg: Any) -> Any:
	"""
	Return an immutable shallow copy
	:param arg:
		dict: ReadOnlyDict is returned
		list: tuple is returned
		set: frozenset is returned
		otherwise: arg is returned 'as is'
	"""
	if isinstance(arg, dict):
		return ReadOnlyDict(
			{
				recursive_freeze(k): recursive_freeze(v)
				for k, v in arg.items()
			}
		)

	if isinstance(arg, list):
		return tuple(
			map(recursive_freeze, arg)
		)

	if isinstance(arg, set):
		return frozenset(arg)

	return arg


def recursive_unfreeze(arg: ReadOnlyDict) -> dict:
	"""
	Inverse of recursive_freeze
	"""
	if isinstance(arg, ReadOnlyDict):
		return dict(
			{
				recursive_unfreeze(k): recursive_unfreeze(v)
				for k, v in arg.items()
			}
		)

	if isinstance(arg, tuple):
		return list(
			map(recursive_unfreeze, arg)
		)

	if isinstance(arg, frozenset):
		return set(arg)

	return arg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/view/T3Store.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                01.12.2021
# Last Modified Date:  16.01.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Container, Iterator, Iterable, Sequence
from ampel.types import JDict, OneOrMany
from ampel.view.T3DocView import T3DocView
from ampel.view.ReadOnlyDict import ReadOnlyDict
from ampel.content.T3Document import T3Document
from ampel.config.AmpelConfig import AmpelConfig
from ampel.util.mappings import dictify
from ampel.util.hash import build_unsafe_dict_id


class T3Store:

	__slots__ = 'views', 'units', 'session', 'extra'

	# Sequence of T3 views (see T3IncludeDirective.docs and T3Processor.include)
	views: None | Sequence[T3DocView]

	# Set of unit names of the t3 views contained in this store
	units: set[str]

	# Read-only dict containing session information available in this store.
	# (see T3IncludeDirective.session and T3Processor.include)
	# Note that session infos are saved into 'meta' and therefore must be BSON encodable
	session: None | JDict

	# Free-form dict (usable for dedicated inter-units communication for example)
	extra: JDict


	@classmethod # Static ctor
	def of(cls, docs: None | Sequence[T3Document], conf: AmpelConfig) -> "T3Store":
		return cls(
			views = tuple(
				T3DocView.of(doc, conf)
				for doc in docs
			) if docs else None
		)


	def __init__(self, views: None | Sequence[T3DocView] = None, session: None | JDict = None):
		object.__setattr__(self, 'views', views)
		object.__setattr__(self, 'session', ReadOnlyDict(session) if session else None)
		object.__setattr__(self, 'units', set(v.unit for v in views) if views else set())
		object.__setattr__(self, 'extra', {})


	def __setattr__(self, k, v):
		raise ValueError("T3Store is read only")


	def __delattr__(self, k):
		raise ValueError("T3Store is read only")


	def add_view(self, t3v: T3DocView) -> None:
		self.units.add(t3v.unit)
		if self.views:
			l = list(self.views) if self.views else []
			l.append(t3v)
			object.__setattr__(self, 'views', tuple(l))
		else:
			object.__setattr__(self, 'views', (t3v, ))


	def add_session_info(self, d: JDict) -> None:
		if self.session:
			object.__setattr__(self, 'session', ReadOnlyDict(self.session | d))
		else:
			object.__setattr__(self, 'session', ReadOnlyDict(d))


	def contains(self, unit: str | Iterable[str]) -> bool:
		if isinstance(unit, str):
			return unit in self.units
		return bool(len(set(unit) - self.units))


	def get_mandatory_view(self,
		unit: None | str | Container[str] = None, *,
		config: None | int | JDict = None,
		code: None | int = None
	) -> T3DocView:
		if (x := self.get_view(unit, config=config, code=code)):
			return x
		raise ValueError(f"{unit} results are required")

			
	def get_view(self,
		unit: None | str | Container[str] = None, *,
		config: None | int | JDict = None,
		code: None | int = None
	) -> None | T3DocView:
		return next(
			self.get_views(unit, config, code),
			None
		)


	def get_views(self,
		unit: str | Container[str] | None = None,
		config: None | OneOrMany[int | JDict] = None,
		code: None | int = None
	) -> Iterator[T3DocView]:
		"""
		Get a subset of T3 documents.

		:param unit: limits the returned science record(s) to the one with the provided t2 unit id
		"""

		if not self.views:
			return None

		units: None | Container[str] = [unit] if isinstance(unit, str) else unit

		if config is None:
			configs = None
		elif isinstance(config, dict):
			configs = [build_unsafe_dict_id(dictify(config))]
		elif isinstance(config, int):
			configs = [config]
		elif isinstance(config, Sequence):
			configs = [
				el if isinstance(el, int) else build_unsafe_dict_id(dictify(el))
				for el in config
			]
		else:
			configs = config

		for t3v in self.views:
			if units and t3v.unit not in units:
				continue
			if configs and t3v.confid not in configs:
				continue
			if code is not None and t3v.code != code:
				continue
			yield t3v



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/abstract/AbsT3PlainUnit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                08.12.2021
# Last Modified Date:  12.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.types import UBson
from ampel.view.T3Store import T3Store
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.LogicalUnit import LogicalUnit
from ampel.struct.UnitResult import UnitResult


class AbsT3PlainUnit(AmpelABC, LogicalUnit, abstract=True):
	""" Generic abstract class for T3 units receiving only a T3Store via the process method """

	@abstractmethod
	def process(self, t3s: T3Store) -> UBson | UnitResult:
		"""
		The content of the t3 store is dependent on:
		- the configuration of the 'include' option of the underlying t3 process
		- previously run t3 units if the option 'propagate' is activated
		"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/base/AlternativeAmpelBaseModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                07.10.2019
# Last Modified Date:  30.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from __future__ import annotations
import collections.abc as abc
from ujson import loads, dumps # type: ignore
from typeguard import check_type # type: ignore
from types import MemberDescriptorType, GenericAlias, UnionType
import ampel.types as altypes
from ampel.types import Traceless, TRACELESS
from ampel.secret.Secret import Secret
from typing import Any, Type, Union, Annotated, get_origin, get_args, _GenericAlias, _UnionGenericAlias # type: ignore[attr-defined]

ttf = type(Traceless)
NoneType = type(None)


class AlternativeAmpelBaseModel:
	"""
	Prototype class, not in used.
	This class supports setting slots values through constructor parameters (they will be type checked as well).
	Type checking can be deactivated globally by setting ampel.types.do_type_check to False
	"""

	_annots: dict[str, Any] = {}
	_defaults: dict[str, Any] = {}
	_slot_defaults: dict[str, Any] = {}
	_aks: set[str] = set() # annotation keys
	_sks: set[str] = set() # slots keys
	_debug: int = 0


	@classmethod
	def __init_subclass__(cls, *args, **kwargs) -> None:
		"""
		Combines annotations & default values of this class with the one defined in sub-classes.
		Has similarities with the newly introduced typing.get_type_hints() function.
		"""
		super().__init_subclass__(*args, **kwargs) # type: ignore

		joined_ann = {
			k: v for k, v in cls._annots.items()
			if not (k[0] == '_' or 'ClassVar' in str(v))
		}
		joined_defaults = cls._defaults.copy()
		joined_sks = cls._sks.copy()
		joined_aks = cls._aks.copy()

		for base in reversed(cls.mro()):
			if ann := getattr(base, '__annotations__', {}):
				defs = getattr(base, '__dict__', {})
				for k, v in ann.items():
					if k == '__slots__' or k[0] == '_' or 'ClassVar' in str(v):
						continue
					joined_ann[k] = v # update merged annotations
					joined_aks.add(k) # update set of known attribute names
					if k in defs:
						if type(defs[k]) is MemberDescriptorType: # is a slot
							if k in cls._slot_defaults:
								joined_defaults[k] = cls._slot_defaults[k]
							continue
						joined_defaults[k] = base.__dict__[k]
					# if None |  with no default
					elif get_origin(v) in (Union, UnionType) and NoneType in get_args(v) and k not in joined_defaults: # type: ignore[misc]
						joined_defaults[k] = None
					elif k in cls._slot_defaults:
						joined_defaults[k] = cls._slot_defaults[k]

			# allow subclasses to change default value without supplying a new annotation
			for k, v in getattr(base, '__dict__', {}).items():
				if k in joined_ann and not k in ann:
					joined_defaults[k] = v

		# Coercion (models, secrets) and type checking if ampel.types.do_type_check is True
		for k, v in joined_defaults.items():
			joined_defaults[k] = cls.get_value(k, joined_ann[k], v)

		if slots := getattr(cls, '__slots__', None):
			joined_sks.update(slots)

		for el in (
			('_annots', joined_ann), ('_defaults', joined_defaults),
			('_aks', joined_aks), ('_sks', joined_sks), ('_model', None)
		):
			setattr(cls, el[0], el[1])


	@classmethod
	def validate(cls, value: dict, _omit_traceless: bool = True) -> Any:
		""" Validate kwargs values against the fields of cls """

		defs = cls._defaults
		for k, v in cls._annots.items():
			if type(v) is ttf and v.__metadata__[0] == TRACELESS:
				continue
			if k in value:
				if cls._has_nested_model(v):
					v = cls._modelify(k, v, value[k])[1]
				if isinstance(v, type) and issubclass(v, Klass):
					v.validate(value[k], _omit_traceless=_omit_traceless)
				else:
					check_type(k, value[k], v)
			elif k in defs:
				continue
			else:
				raise ValueError(f"Parameter required: {k}")

		for k in value.keys():
			if k not in cls._annots:
				raise ValueError(f"Unknown parameter {k}")


	@classmethod
	def _spawn_model(cls, model: "type[AlternativeAmpelBaseModel]", value: dict) -> "AlternativeAmpelBaseModel":

		args = {
			k: model._spawn_model(model, v) if (
				k in cls._annots and
				isinstance(cls._annots[k], type) and
				issubclass(cls._annots[k], cls)
			) else v
			for k, v in value.items()
		}

		if cls._debug > 0:
			print(f"Spawning {model.__name__} with: {args}")

		try:
			ret = model(**args)
		except Exception as e:
			if cls._debug > 0:
				print(f"{model.__name__} instantiation failed")
				print("##", repr(e))
				import traceback
				print(traceback.format_exc())
			raise e

		if cls._debug > 0:
			print(f"{model.__name__} spawned")

		return ret


	@classmethod
	def _has_nested_model(cls, annot: type) -> bool:

		if cls._debug > 1:
			print("Checking " + str(annot).replace('collections.abc.', '') + " for nested models")

		if isinstance(annot, _GenericAlias) and hasattr(annot, 'mro') and issubclass(get_origin(annot), Secret): # type: ignore
			return True

		for i, el in enumerate(get_args(annot)):

			if cls._debug > 1:
				print(f"get_args element {i}: ", el)

			if isinstance(el, (GenericAlias, _GenericAlias, _UnionGenericAlias, UnionType)):
				o = get_origin(el)
				if (isinstance(o, type) and issubclass(o, Klass)) or cls._has_nested_model(el):
					return True

			if isinstance(el, type) and issubclass(el, Klass):
				return True

		return False


	@classmethod
	def _modelify(cls, key: str, annot: type, arg: Any) -> tuple[Exception | bool, Any]:

		oa = get_origin(annot)

		if cls._debug:
			print(f"● Potentially spawning model for {cls.__name__}.{key}")
			print("Annotation: ", annot)
			if oa:
				print("Origin", oa)
			print("Value:", arg)

		if oa is Union or oa is UnionType:
			if cls._debug:
				print("Origin is Union")
			es: list[Exception] = []
			for a in get_args(annot):
				try:
					if cls._debug > 1:
						print("Considering union member: ", a)
					modified, ret = cls._modelify(key, a, arg)
					if modified is True:
						if cls._debug > 1:
							print("Model spawned by union member: ", a)
						return True, ret
					elif isinstance(modified, Exception):
						es.append(modified)
				except Exception as e:
					es.append(e)
			if es and cls._has_nested_model(annot):
				raise TypeError(
					f"Type of {cls.__name__}.{key} must be one of {str(annot).replace('typing.', '')}:\n" +
					f"Value: {arg}\nAssociated errors:\n• " +
					"\n• ".join([str(e) for e in es])
				)
			return False, arg

		elif oa is Annotated:
			return cls._modelify(key, get_args(annot)[0], arg)
				
		elif oa is dict and isinstance(arg, dict):
			if cls._debug:
				print("Origin is dict")
			modified = False
			ret = {}
			for k, v in arg.items():
				x, ret[k] = cls._modelify(f"{key}.{k}", get_args(annot)[1], v)
				if x is True:
					modified = True
			return modified, ret

		elif oa in (list, tuple, abc.Sequence) and isinstance(arg, (list, tuple)):
			if cls._debug:
				print("Origin is sequence")
			ret = []
			modified = False
			for i, el in enumerate(arg):
				x, y = cls._modelify(f"{key}[{i}]", get_args(annot)[0], el)
				ret.append(y)
				if x is True:
					modified = True
			return modified, ret

		elif oa is None and isinstance(arg, dict) and isinstance(annot, type) and issubclass(annot, AlternativeAmpelBaseModel):
			try:
				return True, cls._spawn_model(annot, arg)
			except Exception as e:
				return e, arg

		elif isinstance(oa, type) and issubclass(oa, Klass) and isinstance(arg, dict):
			try:
				return True, cls._spawn_model(oa, arg)
			except Exception as e:
				return e, arg

		else:
			if cls._debug > 1:
				print("Value left as is")

		return False, arg


	@classmethod
	def get_value(cls, k: str, a: type, v: Any) -> Any:

		es = None
		ao = get_origin(a)
		if a in (bool, int, str, float):
			pass
		elif cls._has_nested_model(a):
			if cls._debug > 1:
				print(f"{cls.__name__}.{k} has nested model")
			es, v = cls._modelify(k, a, v)
		elif ao in (Union, UnionType):
			if cls._debug > 1:
				print(f"{cls.__name__}.{k} has Union type")
			es, v = cls._modelify(k, a, v)
		elif ao is Annotated:
			return cls.get_value(k, get_args(a)[0], v)
		elif isinstance(a, type) and issubclass(a, Klass) and isinstance(v, dict):
			v = cls._spawn_model(a, v)
		elif isinstance(a, type) and issubclass(a, Secret):
			v = a(**v)
		else:
			if cls._debug > 1:
				print(f"{cls.__name__}.{k} has no nested model")

		if ao is Annotated:
			return cls.get_value(k, get_args(a)[0], v)

		if altypes.do_type_check:
			"""
			from typingx import isinstancex
			if cls._debug > 1:
				print(f"○ Type checking {cls.__name__}.{k}\n  Annotated type: {a}\n  Value: {v}")
			if isinstancex(v, a):
				if cls._debug > 1:
						print("✓ OK")
			else:
				msg = f"{cls.__name__}: invalid type for field '{k}'\n  Type: {a}\n  Value: {v}"
				if isinstance(es, list):
					msg += "Related errors:\n" + "\n".join([str(e) for e in es])
				raise TypeError(msg) from None
			"""
			try:
				if cls._debug > 1:
					print(f"○ Type checking {cls.__name__}.{k}\n  Annotated type: {a}\n  Value: {v}")
				check_type(k, v, a)
				if cls._debug > 1:
					print("✓ OK")
			except Exception as e:
				msg = f"{str(e)}\nField: {cls.__name__}.{k}\nType: {a}\nValue: {v}"
				if isinstance(es, list):
					msg += "Related errors:\n" + "\n".join([str(e) for e in es])
				raise TypeError(msg) from None

		return v


	def __init__(self, **kwargs) -> None:

		cls = self.__class__
		sa = self.__setattr__
		defs = cls._defaults
		self._exclude_unset: set[str] = set()

		if cls._debug > 1:
			print(f"{cls.__name__}.__init__(...)")

		for k, a in cls._annots.items():

			if cls._debug > 1:
				print(f"□ {cls.__name__}.{k}: {a}")

			es = None

			# Should be first
			if k in kwargs:
				v = cls.get_value(k, a, kwargs[k])
			elif k in defs:
				self._exclude_unset.add(k)
				if isinstance(defs[k], (list, dict)):
					v = loads(dumps(defs[k]))
				elif Klass in defs[k].__class__.__mro__:
					v = defs[k].__class__(**loads(dumps(defs[k].dict())))
				else:
					v = defs[k]
			elif k in cls._slot_defaults:
				self._exclude_unset.add(k)
				if isinstance(cls._slot_defaults[k], (list, dict)):
					v = loads(dumps(cls._slot_defaults[k]))
				elif Klass in cls._slot_defaults[k].__class__.__mro__:
					v = cls._slot_defaults[k].__class__(**loads(dumps(cls._slot_defaults[k].dict())))
				else:
					v = cls._slot_defaults[k]
			else:
				raise ValueError(f"{cls.__name__}: a value is required for parameter '{k}'")

			if cls._debug > 1:
				print(f"Setting {cls.__name__}.{k} value: {v}")

			sa(k, v)

		for k in kwargs.keys():
			if k not in cls._annots:
				raise ValueError(f"{cls.__name__}: unknown parameter {k}")

		if cls._debug > 1:
			print(f"End of {cls.__name__}.__init__(...)")


	def dict(self,
		include: None | set[str] = None,
		exclude: None | set[str] = None,
		exclude_defaults: bool = False,
		exclude_unset: bool = False,
	) -> dict[str, Any]:

		d = self.__dict__
		incl = self._aks if include is None else include
		excl = {
			k for k, v in self._annots.items()
			if type(v) is ttf and v.__metadata__[0] == TRACELESS
		}

		if exclude is not None:
			excl.update(exclude)

		if exclude_unset:
			excl.update(self._exclude_unset)

		if exclude_defaults:
			for k in self._defaults:
				if k in self.__dict__:
					if self.__dict__[k] == self._defaults[k]:
						excl.add(k)
				else:
					excl.add(k)

		return {
			k: self._dictify(v)
			for k, v in d.items()
			if k in incl and not (k in excl or isinstance(d[k], Secret))
		}


	def _dictify(self, arg: Any, dict_kwargs={}) -> Any:
		if isinstance(arg, (list, tuple, set)):
			return [self._dictify(el, dict_kwargs) for el in arg]
		elif isinstance(arg, dict):
			return {
				k: self._dictify(v, dict_kwargs)
				for k, v in arg.items()
			}
		elif isinstance(arg, Klass):
			return arg.dict(**dict_kwargs)
		return arg


Klass = AlternativeAmpelBaseModel



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/abstract/AbsIdMapper.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                12.02.2021
# Last Modified Date:  12.02.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import overload
from ampel.types import StrictIterable, StockId
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.AmpelBaseModel import AmpelBaseModel


class AbsIdMapper(AmpelABC, AmpelBaseModel, abstract=True):

	@overload
	@classmethod
	def to_ampel_id(cls, ext_id: str) -> int:
		...

	@overload
	@classmethod
	def to_ampel_id(cls, ext_id: StrictIterable[str]) -> list[int]:
		...

	@classmethod
	@abstractmethod
	def to_ampel_id(cls, ext_id: str | StrictIterable[str]) -> int | list[int]:
		...

	@overload
	@classmethod
	def to_ext_id(cls, ampel_id: StockId) -> str:
		...

	@overload
	@classmethod
	def to_ext_id(cls, ampel_id: StrictIterable[StockId]) -> list[str]:
		...

	@classmethod
	@abstractmethod
	def to_ext_id(cls, ampel_id: StockId | StrictIterable[StockId]) -> str | list[str]:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/base/BadConfig.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.02.2021
# Last Modified Date:  17.02.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

class BadConfig(Exception):
	pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/abstract/AbsTiedStockT2Unit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.02.2021
# Last Modified Date:  03.04.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Union
from collections.abc import Sequence
from ampel.types import UBson
from ampel.struct.UnitResult import UnitResult
from ampel.view.T2DocView import T2DocView
from ampel.base.decorator import abstractmethod
from ampel.content.StockDocument import StockDocument
from ampel.abstract.AbsTiedT2Unit import AbsTiedT2Unit


class AbsTiedStockT2Unit(AbsTiedT2Unit, abstract=True):
	""" Later """

	@abstractmethod
	def process(self, stock: StockDocument, t2_view: Sequence[T2DocView]) -> UBson | UnitResult:
		"""
		Returned object should contain computed science results to be saved into the DB.

		.. note:: the returned dict must have only string keys and be BSON-encodable
		"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/content/MetaRecord.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.05.2021
# Last Modified Date:  07.10.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Literal, Any, TypedDict
from collections.abc import Sequence
from ampel.types import ChannelId, Tag
from ampel.content.MetaActivity import MetaActivity


class MetaRecord(TypedDict, total=False):
	""" A record of updates on tier documents """

	#: Run(s) associated with this record
	run: int

	#: UNIX epoch of the activity
	ts: int | float

	#: Tier of the associated process
	tier: Literal[-1, 0, 1, 2, 3]

	#: Channels associated with the activity
	channel: ChannelId | Sequence[ChannelId]

	#: Trace ids associated with the creation / update of this document
	traceid: dict[str, None | int]

	#: hash of potentially underlying job schema
	jobid: None | int

	#: Name of the associated process
	process: int | str

	#: Free-form labels
	tag: Tag | Sequence[Tag]

	#: Status code of the latest associated process
	code: int

	#: Actions performed on the associated document
	activity: Sequence[MetaActivity]

	#: Duration of the process
	duration: int | float

	#: Free-form information
	extra: dict[str, Any]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/content/DataPoint.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                12.12.2019
# Last Modified Date:  25.06.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, TypedDict
from typing_extensions import Required
from collections.abc import Sequence
from ampel.types import StockId, DataPointId, ChannelId
from ampel.content.MetaRecord import MetaRecord


class DataPoint(TypedDict, total=False):
	"""
	A single data point.

	A dict containing 1 or more of the following items:
	"""

	id: Required[DataPointId]
	stock: StockId | Sequence[StockId]
	origin: int
	tag: Sequence[int | str]
	channel: Required[Sequence[ChannelId]]
	meta: Required[Sequence[MetaRecord]]
	excl: Sequence[ChannelId]
	body: Required[dict[str, Any]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/model/DPSelection.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                10.03.2020
# Last Modified Date:  28.09.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Literal
from collections.abc import Callable
from ampel.model.UnitModel import UnitModel
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.abstract.AbsApplicable import AbsApplicable
from ampel.base.AmpelBaseModel import AmpelBaseModel


class DPSelection(AmpelBaseModel):

	#: Aux unit name used to filter datapoints.
	# (Unit must define method apply(self, arg: Any) -> Any)
	filter: None | str | UnitModel[str] = None

	#: Dict key used for sorting
	#: Use 'id' to sort dps based on key 'id' at root level,
	#: anything else will target a field (by name) in 'body'
	sort: None | str = None

	#: How to slice the datapoint array.
	#: If a list is provided, it must have a length of 3
	select: None | Literal['first', 'last'] | tuple[None|int, None|int, None|int] = None


	def __init__(self, **kwargs):

		if 'select' in kwargs and isinstance(kwargs['select'], list) and len(kwargs['select']) == 3:
			kwargs['select'] = tuple(kwargs['select'])

		super().__init__(**kwargs)

		if self.sort and not self.select:
			raise ValueError("Options 'sort' requires option 'select'")


	def tools(self) -> tuple[None|AbsApplicable, None|Callable[[list], list], slice]:

		so = None
		sl = None

		f = AuxUnitRegister.new_unit(
			self.filter if isinstance(self.filter, UnitModel) else UnitModel(unit=self.filter),
			sub_type = AbsApplicable
		) if self.filter else None

		if self.sort:
			if self.sort == 'id':
				lbd = lambda k: k['id']
			else:
				lbd = lambda k: k['body'][self.sort]
			so = lambda l: sorted(l, key=lbd)

		if self.select is None:
			sl = slice(None)
		elif self.select == "all":
			sl = slice(None)
		elif self.select == "first":
			sl = slice(1)
		elif self.select == "last":
			sl = slice(-1, -2, -1)
		elif isinstance(self.select, (list, tuple)) and len(self.select) == 3:
			sl = slice(*self.select)
		else:
			raise ValueError(
				f"Unsupported value provided as slice parameter : {self.select}"
			)

		return f, so, sl



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/content/LogDocument.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.02.2020
# Last Modified Date:  16.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import sys
if sys.version_info.minor > 8:
	from typing import TypedDict
else:
	from typing_extensions import TypedDict
from typing import Any
from collections.abc import Sequence
from ampel.types import ChannelId, StockId


class ChannelLogEntry(TypedDict):
	"""
	Abbreviations:
	s: stock, a: alert, f: flag, r: run, m: msg, c: channel

	Used to save multiple channel specific messages into a single LogDocument
	Example of a LogDocument embedding two ChannelLogEntry entries: {
		"_id" : ObjectId("5be4aa6254048041edbac352"),
		"s" : NumberLong(1810101032122523),
		"a" : NumberLong(404105201415015004),
		"f" : 572784643,
		"r" : 509,
		"m" : [
			{"c" : "NO_FILTER", "m": "Alert accepted"},
			{"c" : "HU_RANDOM", "m": "Alert accepted"},
		]
	}
	"""
	#: channel
	c: ChannelId

	#: msg
	m: str


class LogDocument(TypedDict, total=False):
	"""
	Abbreviations:
	s: stock, a: alert, f: flag, r: run, m: msg, c: channel

	Example: {
		"_id" : ObjectId("5be4aa6254048041edbac353"),
		"s" : NumberLong(1810101032122523),
		"a" : NumberLong(404105201415015004),
		"f" : 572784643,
		"r" : 509,
		"c" : "NO_FILTER",
		"m" : "Alert accepted"
	}
	"""

	#: database key
	_id: bytes

	#: flag
	f: int

	#: run id
	r: int | Sequence[int]

	#: msg
	m: str | Sequence[str] | ChannelLogEntry

	#: stock
	s: StockId | Sequence[StockId]

	#: channel
	c: ChannelId | Sequence[ChannelId]

	#: extra
	e: dict[str, Any]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/abstract/AbsCustomStateT2Unit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                28.12.2019
# Last Modified Date:  03.04.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Generic
from collections.abc import Iterable
from ampel.types import T, UBson
from ampel.struct.UnitResult import UnitResult
from ampel.content.T1Document import T1Document
from ampel.content.DataPoint import DataPoint
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.LogicalUnit import LogicalUnit


class AbsCustomStateT2Unit(Generic[T], AmpelABC, LogicalUnit, abstract=True):
	"""
	A T2 unit bound to a custom type *constructed* from a :class:`compound <ampel.content.T1Document.T1Document>`

	Known subclass: :class:`~ampel.abstract.AbsLightCurveT2Unit.AbsLightCurveT2Unit`
	"""

	# We want to enforce the implementation of an abstract *class method*
	# and hence have purposely omitted the first reflective argument
	@staticmethod
	@abstractmethod(force=True)
	def build(compound: T1Document, datapoints: Iterable[DataPoint]) -> T:
		"""
		Create the parametrized type using compound and datapoints.
		For example, AbsCustomStateT2Unit[LightCurve] would return a
		:class:`~ampel.view.LightCurve.LightCurve` instance.
		"""
		...


	@abstractmethod
	def process(self, arg: T) -> UBson | UnitResult:
		"""
		Returned object should contain computed science results to be saved into the DB.

		.. note:: the returned dict must have only string keys and be BSON-encodable
		"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/secret/AmpelVault.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                21.06.2021
# Last Modified Date:  22.06.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, overload
from ampel.abstract.AbsSecretProvider import AbsSecretProvider
from ampel.secret.Secret import Secret, T
from ampel.secret.NamedSecret import NamedSecret


class AmpelVault:
	""" Collection of secret providers """

	def __init__(self, providers: list[AbsSecretProvider]) -> None:
		self.providers = providers

	def resolve_secret(self, secret: Secret, ValueType: type) -> bool:

		for sp in self.providers:
			if sp.tell(secret, ValueType):
				return True
		return False

	@overload
	def get_named_secret(self, label: str) -> None | NamedSecret[Any]:
		...
	
	@overload
	def get_named_secret(self, label: str, ValueType: type[T]) -> None | NamedSecret[T]:
		...
	
	def get_named_secret(self, label, ValueType=object):
		""" Returns a resolved NamedSecret using provided label """
		ns: NamedSecret = NamedSecret(label=label)
		if self.resolve_secret(ns, ValueType):
			return ns
		return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/config/AmpelConfig.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                22.10.2019
# Last Modified Date:  04.08.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import yaml, json
from typing import Union, Literal, Any, TypeVar, overload, get_origin
from ampel.util.freeze import recursive_freeze
from ampel.view.ReadOnlyDict import ReadOnlyDict

UJson = Union[None, str, int, float, bool, list[Any], dict[str, Any]]
JT = TypeVar('JT', None, str, int, float, bool, bytes, list[Any], dict[str, Any])


class AmpelConfig:
	"""Container for the central Ampel configuration"""

	_config: dict
	_check_types: int = 1


	@classmethod
	def load(cls, config_file_path: str, freeze: bool = True) -> 'AmpelConfig':
		with open(config_file_path, "r") as f:
			return cls(yaml.safe_load(f), freeze)


	def __init__(self, config: dict, freeze: bool = False) -> None:
		"""
		:raises: ValueError if provided config is None or empty
		"""
		if config is None or not config:
			raise ValueError("Please provide a config")

		# Convert potentially stringified int keys (JSON compatibility) back to int
		for s in ('channel', 'confid'):
			for k in [el for el in config[s].keys() if isinstance(el, str) and el.isdigit()]:
				config[s][int(k)] = config[s].pop(k)

		self._config: dict = recursive_freeze(config) if freeze else config

		if 'general' in config and 'check_types' in config['general']:
			self._check_types = config['general']['check_types']

	# Overloads for method call without 'entry'

	@overload
	def get(self) -> dict[str, Any]:
		""" config.get() """

	@overload
	def get(self, entry: None) -> dict[str, Any]:
		""" config.get(None) """

	@overload
	def get(self, entry: None, ret_type: Any) -> dict[str, Any]:
		""" config.get(None, None/dict) """

	@overload
	def get(self, entry: None, ret_type: Any, *, raise_exc: bool) -> dict[str, Any]:
		""" config.get(None, None/dict, raise_exc=True/False) """

	@overload
	def get(self, entry: None, *, raise_exc: bool) -> dict[str, Any]:
		""" config.get(None, raise_exc=True/False) """


	# Overloads for method call with 'entry' but without return type

	@overload
	def get(self, entry: str | list[str]) -> UJson:
		""" config.get('logging') """

	@overload
	def get(self, entry: str | list[str], ret_type: None) -> UJson:
		""" config.get('logging', None) """

	@overload
	def get(self, entry: str | list[str], *, raise_exc: bool) -> UJson:
		""" config.get('logging', raise_exc=False/True) """

	@overload
	def get(self, entry: str | list[str], ret_type: None, *, raise_exc: bool) -> UJson:
		""" config.get('logging', None, raise_exc=False/True) """


	# Overloads for method call with 'entry' and return type

	@overload
	def get(self, entry: str | list[str], ret_type: type[JT]) -> None | JT:
		""" config.get('logging', dict) """

	@overload
	def get(self, entry: str | list[str], ret_type: type[JT], *, raise_exc: Literal[False]) -> None | JT:
		""" config.get('logging', dict, raise_exc=False) """

	@overload
	def get(self, entry: str | list[str], ret_type: type[JT], *, raise_exc: Literal[True]) -> JT:
		""" config.get('logging', dict, raise_exc=True) """


	def get(self, # type: ignore[misc]
		entry: None | str | list[str] = None,
		ret_type: None | type[JT] = None,
		*, raise_exc: bool = False
	) -> UJson | None | JT:
		"""
		Optional arguments:
		
		:param ret_type: expected return type (str, int, dict, list, ...).
		:param entry: sub-config element will be returned.
		
		Examples::
			
			get("channel.HU_RANDOM")
		
		::
			
			get(['foo', 'bar', 'baz'])
		
		:raise ValueError: if the retrieved value has not the expected type
		"""

		if entry is None:
			return self._config

		if isinstance(entry, str):
			entry = entry.split(".")

		ret = self._config # pointer

		# Integerizes int path elements encoded as str
		for el in [(el if not el.isdigit() else int(el)) for el in entry]:
			if el not in ret:
				if raise_exc:
					raise ValueError(f'Config element \'{".".join(entry)}\' not found')
				return None
			ret = ret[el]

		if ret_type:

			if origin := get_origin(ret_type):
				ret_type = origin

			if not isinstance(ret, ret_type): # type: ignore[arg-type]
				raise ValueError(
					f"Retrieved value has not the expected type.\n"
					f"Expected: {ret_type}\n"
					f"Found: {type(ret)}"
				)

		return ret


	def get_conf_id(self, conf_id: int) -> dict[str, Any]:

		if conf_id not in self._config['confid']:
			raise ValueError(f"Config with id {conf_id} not found")

		return self._config['confid'][conf_id]


	def print(self,
		entry: None | str = None, format: Literal['json', 'yaml'] = 'yaml'
	) -> None:

		out = self.get(entry)
		print(
			yaml.dump(out) if format == 'yaml'
			else json.dumps(out, indent=4)
		)


	def freeze(self) -> None:
		if not self.is_frozen():
			self._config = recursive_freeze(self._config)


	def is_frozen(self) -> bool:
		return isinstance(self._config, ReadOnlyDict)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/abstract/AbsCLIOperation.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.03.2021
# Last Modified Date:  22.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from collections.abc import Sequence
from argparse import ArgumentParser
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser


class AbsCLIOperation(AmpelABC, abstract=True):
	"""
	Implementing subclasses shall be placed in package ampel.cli
	"""

	@abstractmethod
	def get_parser(self, sub_op: None | str = None) -> ArgumentParser | AmpelArgumentParser:
		...

	@abstractmethod
	def run(self, args: dict[str, Any], unknown_args: Sequence[str], sub_op: None | str = None) -> None:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/cli/LoadAnyOfAction.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                18.03.2021
# Last Modified Date:  22.10.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from argparse import Action
from ampel.model.operator.AnyOf import AnyOf

class LoadAnyOfAction(Action):

	def __call__(self, parser, namespace, values, option_string=None):

		v = [
			int(el) if el.lstrip("-+").isdigit() else el
			for el in values
		]

		setattr(
			namespace,
			self.dest,
			v[0] if len(v) == 1 else AnyOf(any_of = v)
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsApplicable.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                27.02.2020
# Last Modified Date:  09.05.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.AmpelBaseModel import AmpelBaseModel

class AbsApplicable(AmpelABC, AmpelBaseModel, abstract=True):

	@abstractmethod
	def apply(self, arg: Any) -> Any:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/view/ReadOnlyDict.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                Unspecified
# Last Modified Date:  16.04.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>


class ReadOnlyDict(dict):
	"""A dict whose items can't be changed."""

	def __readonly__(self, *args, **kwargs):
		""":raises RuntimeError: whenever called"""
		raise RuntimeError("Cannot modify ReadOnlyDict")

	__setitem__ = __readonly__
	__delitem__ = __readonly__
	pop = __readonly__ # type: ignore
	popitem = __readonly__
	clear = __readonly__
	update = __readonly__ # type: ignore
	setdefault = __readonly__ # type: ignore

	del __readonly__

	def __reduce__(self):
		return type(self), (dict(self),)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/abstract/AbsSecretProvider.py
# License:             BSD-3-Clause
# Author:              Jakob van Santen <jakob.van.santen@desy.de>
# Date:                14.08.2020
# Last Modified Date:  21.06.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Type
from ampel.secret.Secret import Secret
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod


class AbsSecretProvider(AmpelABC, abstract=True):
	"""
	Interface to a secret store used to resolve secrets.
	The underlying store may be as simple as a dict loaded from a JSON file
	or a complete key manager like Vault.
	"""

	@abstractmethod
	def tell(self, arg: Secret, ValueType: type) -> bool:
		"""
		Potentially update an initialized Secret instance with
		the actual sensitive information associable with it.
		:returns: True if the Secret was told/resolved or False
		if the provided Secret is either unknown to this secret
		provider, or resolves to a value of the wrong type.
		"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/base/AmpelUnit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                07.10.2019
# Last Modified Date:  09.01.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import collections.abc as abc
from types import MemberDescriptorType, UnionType
from ampel.types import Traceless, TRACELESS
from ampel.secret.Secret import Secret
from ampel.base.AmpelBaseModel import AmpelBaseModel
from pydantic import BaseModel, validate_model, create_model
from typing import Any, Type, Union, get_origin, get_args

ttf = type(Traceless)
NoneType = type(None)


class AmpelUnit:
	"""
	This class supports setting slots values through constructor parameters (they will be type checked as well).
	Type checking can be deactivated globally by setting ampel.types.do_type_check to False
	"""

	_model: Type[BaseModel]
	_annots: dict[str, Any] = {}
	_defaults: dict[str, Any] = {}
	_slot_defaults: dict[str, Any] = {}
	_aks: set[str] = set() # annotation keys
	_sks: set[str] = set() # slots keys


	@classmethod
	def __init_subclass__(cls, *args, **kwargs) -> None:
		"""
		Combines annotations & default values of this class with the one defined in sub-classes.
		Has similarities with the newly introduced typing.get_type_hints() function.
		"""
		super().__init_subclass__(*args, **kwargs) # type: ignore

		joined_ann = {
			k: v for k, v in cls._annots.items()
			if not (k[0] == '_' or 'ClassVar' in str(v))
		}
		joined_defaults = cls._defaults.copy()
		joined_sks = cls._sks.copy()
		joined_aks = cls._aks.copy()
		# Needed for when sub-classes inherit both AmpelUnit and AmpelBaseModel/BaseModel
		field_keys: Any = getattr(cls, '__fields__', set())

		for base in reversed(cls.mro()):
			if ann := getattr(base, '__annotations__', {}):
				defs = getattr(base, '__dict__', {})
				for k, v in ann.items():
					if k == '__slots__' or k[0] == '_' or 'ClassVar' in str(v):
						continue

					joined_ann[k] = v # update merged annotations
					joined_aks.add(k) # update set of known attribute names

					if k in defs:
						if type(defs[k]) is MemberDescriptorType: # is a slot
							if k in cls._slot_defaults:
								joined_defaults[k] = cls._slot_defaults[k]
							continue
						joined_defaults[k] = base.__dict__[k]
					# if None |  with no default
					elif get_origin(v) in (Union, UnionType) and NoneType in get_args(v) and k not in joined_defaults: # type: ignore[misc]
						joined_defaults[k] = None
					elif k in cls._slot_defaults:
						joined_defaults[k] = cls._slot_defaults[k]
					elif k in field_keys:
						fields = cls.__fields__ # type: ignore[attr-defined]
						if fields[k].required is False:
							joined_defaults[k] = fields[k].default


			# allow subclasses to change default value without supplying a new annotation
			for k, v in getattr(base, '__dict__', {}).items():
				if k in joined_ann and k not in ann:
					joined_defaults[k] = v

		if slots := getattr(cls, '__slots__', None):
			joined_sks.update(slots)

		for el in (
			('_annots', joined_ann), ('_defaults', joined_defaults),
			('_aks', joined_aks), ('_sks', joined_sks), ('_model', None)
		):
			setattr(cls, el[0], el[1])


	@classmethod
	def _create_model(cls, omit_traceless: bool = False) -> Type[BaseModel]:

		defs = cls._defaults
		if hasattr(cls, '__fields__'):
			cls.__fields__.clear() # type: ignore[attr-defined]

		if omit_traceless:
			ttf = type(Traceless)
			kwargs = {
				k: (v, defs[k] if k in defs else ...)
				for k, v in cls._annots.items()
				if not (type(v) is ttf and v.__metadata__[0] == TRACELESS)
			}
		else:
			kwargs = {
				k: (v, defs[k] if k in defs else ...)
				for k, v in cls._annots.items()
			}

		return create_model(
			cls.__name__,
			__config__ = AmpelBaseModel.Config,
			**kwargs # type: ignore
		)


	@classmethod
	def validate(cls, value: dict) -> Any:
		""" Validate kwargs values against fields of cls (except traceless) """
		values, fields, errors = validate_model(cls._create_model(True), value)
		if errors:
			raise TypeError(errors) from None
		return values


	@classmethod
	def validate_all(cls, value: dict) -> Any:
		""" Validate kwargs values against all fields of cls """
		if cls._model is None:
			model = cls._model = cls._create_model()
		else:
			model = cls._model
		values, fields, errors = validate_model(model, value)
		if errors:
			raise TypeError(errors) from None
		return values

	
	def __init__(self, **kwargs) -> None:

		cls = self.__class__

		if cls._model is None:
			cls._model = self._create_model()
			# might be needed in the future due to postponed annotations
			# cls._model.update_forward_refs()

		if hasattr(cls, '__fields__'):
			AmpelBaseModel.__init__(self) # type: ignore[arg-type]

		vres = validate_model(cls._model, kwargs) # type: ignore[arg-type]
		self._exclude_unset = self._defaults.keys() - kwargs.keys()

		# pydantic ValidationError
		if e := vres[2]:
			# https://github.com/samuelcolvin/pydantic/issues/784
			print("")
			if kwargs:
				print("#" * 60)
				print("Offending values:")
				for k, v in kwargs.items():
					print(f"{k}: {v}")
				print("#" * 60)
			raise TypeError(e) from None

		# Save coerced values
		kwargs.update(vres[0])

		sa = self.__setattr__
		sks = cls._sks
		aks = cls._aks

		for k, v in cls._slot_defaults.items():
			if k not in kwargs:
				sa(k, v)

		# Set kwargs attributes
		for k in kwargs:
			if k in sks or k in aks:
				sa(k, kwargs[k])


	def _get_trace_content(self) -> dict[str, Any]:

		a = self._annots
		return {
			k: self._dictify(getattr(self, k))
			for k in sorted(self._annots)
			if not (
				(type(a[k]) is ttf and a[k].__metadata__[0] == TRACELESS) or
				isinstance(a[k], Secret)
			)
		}


	# Note: note defining arguments to avoid mypy complains when mixing AmpelUnit with BaseModel
	def dict(self, **kwargs) -> dict[str, Any]:

		d = self.__dict__
		incl = self._aks if (x := kwargs.get('include')) is None else x
		excl = {
			k for k, v in self._annots.items()
			if type(v) is ttf and v.__metadata__[0] == TRACELESS
		}

		if 'exclude' in kwargs:
			excl.update(kwargs['exclude'])

		if kwargs.get('exclude_unset'):
			excl.update(self._exclude_unset)

		if kwargs.get('exclude_defaults'):
			for k in self._defaults:
				if self.__dict__[k] == self._defaults[k]:
					excl.add(k)

		return {
			k: self._dictify(v)
			for k, v in d.items()
			if k in incl and not (k in excl or isinstance(d[k], Secret))
		}


	def _dictify(self, arg: Any, dict_kwargs={}) -> Any:
		if isinstance(arg, (list, tuple, set)):
			return [self._dictify(el, dict_kwargs) for el in arg]
		elif isinstance(arg, dict):
			return {
				k: self._dictify(v, dict_kwargs)
				for k, v in arg.items()
			}
		elif isinstance(arg, (Klass, BaseModel)):
			return arg.dict(**dict_kwargs)
		return arg


	@classmethod
	def get_model_keys(cls) -> abc.KeysView[str]:
		return cls._annots.keys()


Klass = AmpelUnit



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/alert/AmpelAlert.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                26.01.2020
# Last Modified Date:  02.01.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import operator
from typing import Any
from collections.abc import Callable, Sequence
from ampel.types import JDict, StockId, Tag

# Do not enable customizations of operators by sub-classes for now
ops: dict[str, Callable[[str, Any], bool]] = {
	'>': operator.gt,
	'<': operator.lt,
	'>=': operator.ge,
	'<=': operator.le,
	'==': operator.eq,
	'!=': operator.ne,
	'is': operator.is_,
	'is not': operator.is_not,
	'contains': operator.contains,
	'exists': None, # type: ignore
}


class AmpelAlert:
	"""
	Implements AmpelAlertProtocol
	"""

	__slots__ = '_id', '_stock', '_datapoints', '_tag', '_extra'

	def __init__(self,
		id: int, #: unique identifier for this alert
		stock: StockId, #: stock this alert belongs to
		datapoints: Sequence[JDict],
		tag: None | Tag | list[Tag] = None, #: Optional tag associated with this alert
		extra: None | JDict = None #: Optional information associated with this alert
	) -> None:
		sa = object.__setattr__
		sa(self, '_id', id)
		sa(self, '_stock', stock)
		sa(self, '_datapoints', datapoints)
		sa(self, '_tag', tag)
		sa(self, '_extra', extra)

	@property
	def id(self) -> int:
		return self._id # type: ignore[attr-defined]

	@property
	def stock(self) -> StockId:
		return self._stock # type: ignore[attr-defined]

	@property
	def datapoints(self) -> Sequence[JDict]:
		return self._datapoints # type: ignore[attr-defined]

	@property
	def tag(self) -> None | Tag | list[Tag]:
		return self._tag # type: ignore[attr-defined]

	@property
	def extra(self) -> None | JDict:
		return self._extra # type: ignore[attr-defined]

	def __reduce__(self):
		return (
			type(self),
			(self._id, self._stock, self._datapoints, self._tag, self._extra)
		)

	def __setattr__(self, k, v):
		raise ValueError("AmpelAlert is read only")

	def __delattr__(self, k):
		raise ValueError("AmpelAlert is read only")


	def get_values(self,
		key: str,
		filters: None | Sequence[JDict] = None
	) -> list[Any]:
		"""
		Example:
			
			get_values("magpsf")
		"""
		data = self.apply_filter(self.datapoints, filters) if filters else self.datapoints
		return [el[key] for el in data if key in el]


	def get_tuples(self,
		key1: str, key2: str,
		filters: None | Sequence[JDict] = None
	) -> list[tuple[Any, Any]]:
		"""
		Example::
			
			get_tuples("jd", "magpsf")
		"""
		data = self.apply_filter(self.datapoints, filters) if filters else self.datapoints
		return [
			(el[key1], el[key2])
			for el in data if key1 in el and key2 in el
		]


	def get_ntuples(self,
		params: list[str],
		filters: None | Sequence[JDict] = None
	) -> list[tuple]:
		"""
		Example:
			
			get_ntuples(["fid", "jd", "magpsf"])
		"""
		data = self.apply_filter(self.datapoints, filters) if filters else self.datapoints
		return [
			tuple(el[param] for param in params)
			for el in data if all(param in el for param in params)
		]


	def is_new(self) -> bool:
		return len(self.datapoints) == 1


	def apply_filter(self,
		dicts: Sequence[JDict],
		filters: Sequence[JDict]
	) -> Sequence[JDict]:

		if isinstance(filters, dict):
			filters = [filters]
		else:
			if filters is None or not isinstance(filters, (list, tuple)):
				raise ValueError("Parameter 'filters' must be a dict or a sequence of dicts")

		for f in filters:
			attr = f['attribute']
			op = f['operator']
			if op == 'exists':
				if f['value'] is True:
					dicts = [d for d in dicts if attr in d]
				else:
					dicts = [d for d in dicts if attr not in d]
			else:
				dicts = [d for d in dicts if attr in d and ops[op](d[attr], f['value'])]

		return dicts


	def dict(self) -> JDict:
		return {
			'id': self.id,
			'stock': self.stock,
			'datapoints': self.datapoints,
			'extra': self.extra,
		}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/enum/JournalActionCode.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                01.09.2021
# Last Modified Date:  05.09.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from enum import IntFlag

# flake8: noqa (E221)
class JournalActionCode(IntFlag):
	"""
	Potential code of field 'action' from :class:`~ampel.content.JournalRecord.JournalRecord` (62 different possible values)
	"""

	# T0
	T0_ADD_CHANNEL            = 1
	T0_PULL_CHANNEL           = 2
	T0_ADD_TAG                = 4
	T0_PULL_TAG               = 8
	T0_ADD_EXCL               = 64
	T0_PULL_EXCL              = 128

	# T1
	T1_ADD_CHANNEL            = 1<<8
	T1_PULL_CHANNEL           = 1<<9
	T1_ADD_BODY               = 1<<10
	T1_ADD_TAG                = 1<<11
	T1_PULL_TAG               = 1<<12
	T1_SET_CODE               = 1<<13
	T1_EXTRA_META             = 1<<14
	T1_EXTRA_JOURNAL          = 1<<15

	# T2
	T2_ADD_CHANNEL            = 1<<16
	T2_PULL_CHANNEL           = 1<<17
	T2_ADD_BODY               = 1<<18
	T2_ADD_TAG                = 1<<19
	T2_PULL_TAG               = 1<<20
	T2_SET_CODE               = 1<<21
	T2_EXTRA_JOURNAL          = 1<<22
	T2_EXPORT_DOC             = 1<<23
	T2_IMPORT_RESULT          = 1<<24

	# T3
	T3_ADD_DOC                = 1<<25

	# STOCK
	STOCK_ADD_CHANNEL         = 1<<26
	STOCK_PULL_CHANNEL        = 1<<27
	STOCK_BUMP_UPD            = 1<<28
	STOCK_ADD_TAG             = 1<<29
	STOCK_PULL_TAG            = 1<<30
	STOCK_ADD_NAME            = 1<<31
	STOCK_PULL_NAME           = 1<<32
	STOCK_SET_BODY            = 1<<33

	# CLI / OPS
	STOCK_RESET_BODY          = 1<<34
	T1_RESET_CODE             = 1<<35
	T1_RESET_BODY             = 1<<36
	T2_RESET_CODE             = 1<<37
	T2_RESET_BODY             = 1<<38



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/struct/AmpelBuffer.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                31.05.2018
# Last Modified Date:  01.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, TypedDict, Literal
from ampel.types import StockId
from ampel.content.StockDocument import StockDocument
from ampel.content.DataPoint import DataPoint
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document
from ampel.content.LogDocument import LogDocument

# Please update BufferKey on AmpelBuffer udpates
# There is currently unfortunately no way of extracting a Literal out of a TypedDict
BufferKey = Literal['id', 'stock', 'origin', 't0', 't1', 't2', 'logs', 'extra']

class AmpelBuffer(TypedDict, total=False):
	"""
	Content bundle used to build :class:`~ampel.view.SnapView.SnapView`.
	
	This is a dict containing 1 or more of the following items:
	"""
	# Could stock be of type list[StockDocument] to enable hybrid/dual transients ?
	id: StockId
	stock: None | StockDocument
	origin: None | int | list[int]
	t0: None | list[DataPoint]
	t1: None | list[T1Document]
	t2: None | list[T2Document]
	logs: None | list[LogDocument]
	extra: None | dict[str, Any]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/content/EventDocument.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                04.03.2021
# Last Modified Date:  12.07.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import TypedDict, Any, Literal
from collections.abc import Sequence
from ampel.enum.EventCode import EventCode
from ampel.types import ChannelId


class EventDocument(TypedDict, total=False):
	"""
	Specifications for documents saved into the ampel 'event' collection
	"""

	#: Name of the ampel process (may be hashed for performance reasons)
	process: int | str

	#: process version
	version: int

	#: Ever increasing global and unique run identifier
	run: int

	#: hash of potentially underlying job schema
	jobid: None | int

	#: Optional identification of the associated execution layer
	tier: Literal[-1, 0, 1, 2, 3]

	#: clock time duration of the event
	duration: int | float

	#: Optional channel(s) associated with the underlying process
	channel: ChannelId | Sequence[ChannelId]

	#: A member of :class:`~ampel.enum.EventCode.EventCode`
	code: EventCode

	#: Optional extras
	extra: dict[str, Any]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/enum/MetaActionCode.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                01.09.2021
# Last Modified Date:  10.10.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from enum import IntFlag

# flake8: noqa (E221)
class MetaActionCode(IntFlag):
	"""
	Potential code for 'action' field from :class:`~ampel.content.MetaActivity.MetaActivity`
	"""

	UNIT                      = 1
	CLI                       = 2
	EXTRA_META                = 4
	EXTRA_JOURNAL             = 8
	SET_CODE                  = 16
	SET_UNIT_CODE             = 32
	ADD_INGEST_TAG            = 64
	ADD_UNIT_TAG              = 128
	ADD_WORKER_TAG            = 1<<8
	ADD_OTHER_TAG             = 1<<9
	PULL_TAG                  = 1<<10
	ADD_CHANNEL               = 1<<11
	PULL_CHANNEL              = 1<<12
	ADD_BODY                  = 1<<13
	PULL_BODY                 = 1<<14
	RESET_BODY                = 1<<15
	ADD_T1_EXCL               = 1<<16
	BUMP_STOCK_UPD            = 1<<17



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/abstract/AbsPointT2Unit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                28.12.2019
# Last Modified Date:  28.09.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.types import UBson
from ampel.struct.UnitResult import UnitResult
from ampel.content.DataPoint import DataPoint
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.LogicalUnit import LogicalUnit


class AbsPointT2Unit(AmpelABC, LogicalUnit, abstract=True):
	"""
	A T2 unit bound to a :class:`~ampel.content.DataPoint.DataPoint`

	Note that the implementing class can customize the default ingestion behavior
	(which :class:`~ampel.content.DataPoint.DataPoint` to create
	:class:`T2 documents <ampel.content.T2Document.T2Document>` for)
	by defining the class variable 'eligible'
	
	Example::
	  ingest: ClassVar[DPSelection] = DPSelection(select=(1, -2, 5))
	
	will create documents bound to every 5th datapoint starting from the 2nd
	and ending with the 3rd-to-last

	Example::
	   ingest: ClassVar[DPSelection] = DPSelection(filter=UnitModel(unit='SimpleTagFilter', config={'require': ['ZTF_DP']}), sort='jd', select='first'}

	will create documents bound to the first datapoint with tag 'ZTF_DP', in order of body.jd.
	
	select options:
	  - "first": first datapoint for a stock
	  - "last": most recent datapoint for a stock
	  - "all": every datapoint (default)
	  - :class:`tuple`: :class:`slice` of datapoints
	-> see DPSelection docstring for info regarding 'filter' and 'sort'.

	If 'eligible' is not specified, default ingestion will occur:
	a T2 document will be created for each datapoint
	"""

	@abstractmethod
	def process(self, datapoint: DataPoint) -> UBson | UnitResult:
		"""
		Returned object should contain computed science results to be saved into the DB.

		.. note:: the returned dict must have only string keys and be BSON-encodable
		"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/struct/T1CombineResult.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                04.05.2021
# Last Modified Date:  17.06.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.types import DataPointId


class T1CombineResult:
	"""
	Structure potentialy returned by Ampel AbsT1CombineUnit instances
	to customize code or meta of T1Documents.
	"""

	__slots__ = 'code', 'meta', 'dps'

	def __init__(self,
		dps: list[DataPointId],
		code: None | int = None,
		meta: None | dict[str, Any] = None
	) -> None:
		"""
		:param dps: ids of the datapoints to combine
		:param code: customize DocumentCode (ex: DocumentCode.RERUN_REQUESTED)
		:param meta: customize meta dict
		"""

		self.dps = dps
		self.code = code
		self.meta = meta


	def add_meta(self, key, value) -> 'T1CombineResult':

		if self.meta:
			if self.meta.get(key):
				if isinstance(self.meta.get(key), list):
					self.meta[key].append(value)
				else:
					self.meta[key] = [self.meta[key], value]
			else:
				self.meta[key] = value
		else:
			self.meta = {key: value}

		return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/log/handlers/LoggingHandlerProtocol.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                09.05.2020
# Last Modified Date:  11.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, Protocol, runtime_checkable


@runtime_checkable
class LoggingHandlerProtocol(Protocol):

	level: int

	def handle(self, record: Any) -> None:
		...

	def flush(self) -> None:
		...

class AggregatingLoggingHandlerProtocol(LoggingHandlerProtocol):

	aggregate_interval: float

	def break_aggregation(self) -> None:
		...


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/util/hash.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                22.05.2021
# Last Modified Date:  12.07.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import json, xxhash # type: ignore[import]
from typing import TypeVar
from ampel.util.mappings import flatten_dict

HT = TypeVar("HT", int, bytes, str)
xxfunc = {bytes: 'digest', int: 'intdigest', str: 'hexdigest'}

def hash_payload(payload: bytes, ret: type[HT] = int, size: int = -64) -> HT: # type: ignore[assignment]
	"""
	:param ret: return type, can be bytes, str (hex digest) or int
	:param size: xxhash size. Available sizes: 32, 64, 128 bits.
	Convention: a negative value means the resulting int will be converted to signed int
	(mongodb does not support unsigned integers)

	FYI: using an english dict of 466544 words, hashing using the following algs resulted in:
	64 bits:
	0 collision using xxh64

	32 bits:
	23 collisions using xxh32
	25 collisions with 4-bytes truncated sha1,
	27 collisions with pyhash.fnv1a_32,
	34 collisions with 4-bytes truncated hashlib.blake2b[4bytes]

	In []: hasher = getattr(pyhash, 'xx_32')()
	In []: f=open("words.txt", "r").readlines()
	In []: len(f)
	Out[]: 466544
	In []: s=set()
	In []: for el in f:
				s.add(hasher(el))
	In []: len(f) - len(s)
	Out[]: 23
	"""

	x = getattr(xxhash, f'xxh{abs(size)}_{xxfunc[ret]}')(payload)

	# Convert unsigned to signed int if passed 'size' parameter is negative
	if size < 0 and ret == int and x & (1 << (-size-1)): # type: ignore[operator]
		x = x - 2**-size

	return x


def build_unsafe_dict_id(
	dict_arg: None | dict,
	ret: type[HT] = int, # type: ignore[assignment]
	size: int = -64,
	sort_keys: bool = True,
	flatten_list_members: bool = True,
	flatten_lists: bool = True,
	sort_lists: bool = False
) -> HT:
	"""
	:param dict_arg: can be nested, can be None
	:param ret: return type, can be bytes, str (hex digest) or int
	:param size: see hash_payload docstring
	:param alg: hash algorithm (default is xx_64)
	:param sort_keys: see `flatten_dict` docstring
	:param flatten_list_members: see `flatten_dict` docstring
	:param sort_lists: see `flatten_dict` docstring
	:param flatten_lists: see `flatten_dict` docstring

	Examples:
	In []: build_unsafe_dict_id({'a': 1, 'b': 2, 'c': {'b': 3, 'a': [1, 4]}}, size=32)
	Out[]: 2122149373

	In []: build_unsafe_dict_id({'a': 1, 'b': 2, 'c': {'b': 3, 'a': [1, 4]}}, size=32)
	Out[]: 2122149373

	In []: build_unsafe_dict_id({'a': 1, 'b': 2, 'c': {'b': 3, 'a': [1, 4]}}, size=-64)
	Out[]: -8986814508490313900

	In []: build_unsafe_dict_id({'a': 1, 'b': 2, 'c': {'a': ['r', 1, 4], 'b': 3}}, size=32, ret=str)
	Out[]: 'ec94bc00'

	In []: build_unsafe_dict_id({'b': 2, 'a': 1, 'c': {'b': 3, 'a': [4, 'r', 1]}}, size=32, ret=str)
	Out[]: 'ec94bc00'

	In []: build_unsafe_dict_id({'a': 1, 'b': 2, 'c': {'b': 3, 'a': [1, 5]}}, size=32, ret=str)
	Out[]: 'a6cc3e6b'

	In []: build_unsafe_dict_id({'a': 1, 'b': 2, 'c': {'b': 3, 'a': [1, 4]}}, size=128, ret=str)
	Out[]: '4ee95bc895cf7ef0ddac663121b2b911'

	In []: build_unsafe_dict_id({'a': 1, 'b': 2, 'c': {'b': 3, 'a': [1, 4]}}, size=128)
	Out[]: 104891450430532601648229971339559221521

	Note:
	In []: flatten_dict(
		{'a': 1, 'b': 2, 'c': {'b': 3, 'a': [1, {'d': 1, 'a':1}]}},
		flatten_list_members=True, flatten_lists=True, sort_keys=True
	)
	Out[]: {'a': 1, 'b': 2, 'c.a.0': 1, 'c.a.1.a': 1, 'c.a.1.d': 1, 'c.b': 3}
	"""

	if dict_arg is None:
		dict_arg = {}

	return hash_payload(
		bytes(
			json.dumps(
				flatten_dict(
					dict_arg, '.', sort_keys, flatten_list_members,
					sort_lists, flatten_lists
				),
				indent=None, separators=(',', ':')
			),
			"utf8"
		),
		ret = ret,
		size = size
	)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/abstract/AbsStockT2Unit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                28.12.2019
# Last Modified Date:  11.06.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.types import Union, UBson
from ampel.struct.UnitResult import UnitResult
from ampel.content.StockDocument import StockDocument
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.LogicalUnit import LogicalUnit


class AbsStockT2Unit(AmpelABC, LogicalUnit, abstract=True):
	"""
	A T2 unit bound to a :class:`~ampel.content.StockDocument.StockDocument`
	"""

	@abstractmethod
	def process(self, stock_doc: StockDocument) -> UBson | UnitResult:
		"""
		Returned object should contain computed science results to be saved into the DB.

		.. note:: the returned dict must have only string keys and be BSON-encodable
		"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/cli/LoadJSONAction.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                18.03.2021
# Last Modified Date:  18.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import json
from argparse import Action

class LoadJSONAction(Action):
	def __call__(self, parser, namespace, value, option_string=None):
		setattr(namespace, self.dest, json.loads(value))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/abstract/AbsT0Unit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                07.10.2019
# Last Modified Date:  15.05.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.types import StockId
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.LogicalUnit import LogicalUnit
from ampel.content.DataPoint import DataPoint


class AbsT0Unit(AmpelABC, LogicalUnit, abstract=True):
	"""
	A unit that creates :class:`datapoints <ampel.content.DataPoint.DataPoint>` for Ampel

	Before new datapoint are inserted into the database, they are customized (or 'ampelized' if you will),
	in order to later enable the use of short and flexible queries.
	The cutomizations are light, most of the original information is kept.
	For example, in the case of ZiDataPointShaper:
		* The field candid is renamed in id
		* A new field 'tag' is created
		...
	"""

	@abstractmethod
	def process(self, arg: Any, stock: None | StockId = None) -> list[DataPoint]:
		"""
		Convert an external object to Ampel format
		"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/struct/UnitResult.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                04.05.2021
# Last Modified Date:  20.04.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.types import UBson, OneOrMany, Tag
from ampel.struct.JournalAttributes import JournalAttributes


class UnitResult:
	"""
	Structure potentialy returned by Ampel units to customize
	document information such as body, code, tag or the stock journal entry.
	"""

	__slots__ = 'body', 'tag', 'code', 'journal', 'adapter'

	def __init__(self,
		body: UBson = None,
		tag: None | OneOrMany[Tag] = None,
		code: None | int = None,
		journal: None | JournalAttributes = None,
		adapter: None | str = None
	) -> None:
		"""
		:param body: content for the field 'body' of the associated tier document
		:param tag: tag(s) to be added to the content of the field 'tag' of the associated tier document.
		Note that tags are 'public' attributes (they are not channel bound) and as such, they:
		- are visible by any projection
		- can be used for any db queries
		:param code: customize 'code' of the associated tier document:
		- Either a member of `ampel.enum.DocumentCode` such as DocumentCode.RERUN_REQUESTED (negative values)
		- Or any positive integer number (custom defined for a particular ampel unit)
		:param journal: customization of the journal entry associated with the stock document
		:param adapter: optionally specify which subclass of T3UnitResultAdapter should handle/morph
		this structure before its insertion into the database (ex: 'AmpelPlotAdapter')
		"""
		self.body = body
		self.tag = tag
		self.code = code
		self.journal = journal
		self.adapter = adapter



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/base/decorator.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                27.12.2017
# Last Modified Date:  17.05.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from functools import partial
from typing import TypeVar, overload
from collections.abc import Callable

F = TypeVar('F', bound=Callable)


@overload
def abstractmethod(func: F) -> F:
	...

@overload
def abstractmethod(*, check_signature: bool = False, force: bool = False, var_args: bool = False) -> Callable[[F], F]:
	...

def abstractmethod(func=None, check_signature=False, force=False, var_args: bool = False):
	"""
	:param check_signature: checks method signature (parameter names must equal)
	:param force: whether implementation of methods must be checked even if the subclass is abstract as well
	:param var_args: allow variables arguments (default behavior in native python abc)
	"""

	if func is None:
		return partial(
			abstractmethod, check_signature=check_signature, force=force, var_args=var_args
		)

	if check_signature:
		func.check_signature = True

	if force:
		func.force_check = True

	if var_args:
		func.var_args = True

	func.abstract_method = True
	# emulate abc.abstractmethod to inform autodoc of abstract methods
	func.__isabstractmethod__ = True
	return func


@overload
def defaultmethod(func: F) -> F:
	...

@overload
def defaultmethod(*, check_signature: bool = False, check_super_call: bool = False) -> Callable[[F], F]:
	...

def defaultmethod(func=None, check_signature=False, check_super_call=False):

	if func is None:
		return partial(
			defaultmethod, check_signature=check_signature,
			check_super_call=check_super_call
		)

	func.default_method = True

	if check_signature:
		func.check_signature = True

	if check_super_call:
		func.check_super_call = True

	return func



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/cli/AmpelHelpFormatter.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.03.2021
# Last Modified Date:  18.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from argparse import RawTextHelpFormatter
from ampel.cli.LoadAnyOfAction import LoadAnyOfAction
from ampel.cli.LoadAllOfAction import LoadAllOfAction
from ampel.cli.LoadJSONAction import LoadJSONAction


class AmpelHelpFormatter(RawTextHelpFormatter):

	def __init__(self, prog, indent_increment=2, max_help_position=40, width=None):
		return super().__init__(
			prog, indent_increment=indent_increment,
			max_help_position=max_help_position, width=width
		)

	def add_usage(self, usage, actions, groups, prefix=None):
		if prefix is None:
			prefix = 'Usage: '
		return super().add_usage(usage, actions, groups, prefix)


	def _format_args(self, action, default_metavar):
		"""
		Use custom representation arguments requiring 2 values or JSON string.
		Also, add two spaces to the left column by default
		for making room for potential mutual exclusivity symbol ("\u22BB")
		"""
		if isinstance(action, (LoadAllOfAction, LoadAnyOfAction)):
			return "# # ...  "
		if isinstance(action, LoadJSONAction):
			return "'{#}'  "
		return super()._format_args(action, default_metavar) + "  "

	
	def _format_action(self, action):

		ret = super()._format_action(action).replace(" --", " -").replace("\n ", "\n")

		try:
			if ret.startswith("  {"):
				# Remove {action1,action2,action3} line from help
				return "\n".join(ret.split("\n")[1:])
		except Exception:
			pass

		try:
			if (
				'_mutually_exclusive_groups' in (x := action.__dict__['container'].__dict__) and
				x['_mutually_exclusive_groups'] and
				action.metavar != "@"
			):
				for el in x['_mutually_exclusive_groups']:
					if action in el._group_actions:
						insert_pos = len(ret.split('\n')[0])-len(action.help.split('\n')[0]) - 2
						return ret[0:insert_pos] + "\u22BB" + ret[insert_pos+1:]
						# return "\u22BB" + ret[1:]
		except Exception:
			pass

		return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/content/T3Document.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                23.02.2021
# Last Modified Date:  25.06.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import TypedDict, Any
from typing_extensions import Required
from collections.abc import Sequence
from ampel.types import ChannelId, StockId, Tag, UBson, UnitId
from ampel.content.MetaRecord import MetaRecord


class T3Document(TypedDict, total=False):
	"""
	Specifications for tier3 documents stored as BSON structures in the ampel DB
	"""

	#: Database primary id
	_id: str | bytes

	#: Name of the associated T3 process (may be a hash)
	process: int | str

	#: T3 unit name
	unit: Required[UnitId]

	#: Hash of config (including defaults defined in unit class)
	confid: Required[int]

	#: Resolved T3 unit config for convenience
	config: dict[str, Any]

	#: visible by any projection (not channel bound)
	tag: Tag | Sequence[Tag]

	#: Ampel channel(s) associated with this document
	channel: ChannelId | Sequence[ChannelId]

	#: Note: might contain versions of dependent external services
	#: (not only those used by t3 units but also potentially versions of resources
	#: used by session and/or complement routines)
	meta: Required[MetaRecord]

	#: Records session info, if any
	session: None | dict[str, Any]

	#: Stock id of the views provided to unit
	stock: None | Sequence[StockId]

	#: Optional source origin (avoids potential stock collision between different data sources)
	origin: int

	#: Negative values must be member of :class:`~ampel.enum.DocumentCode.DocumentCode`
	code: Required[int]

	#: Optional human friendly date time stamp
	datetime: str

	#: Optional in t1 and t3 docs, mandatory in t0 and t2 docs
	body: UBson



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/cli/MaybeIntAction.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.03.2021
# Last Modified Date:  18.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from argparse import Action

class MaybeIntAction(Action):

	def __call__(self, parser, namespace, values, option_string=None):
		v = [int(el) if el.lstrip("-+").isdigit() else el for el in values]
		setattr(namespace, self.dest, v[0] if len(v) == 1 else v)
